#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2021 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Set of methods to format data for gemmi parser
"""


def preprocess_cif_category(cif_block, label):
    """
    Checks if the category is present in gemmi.cif.Block object

    Args:
        cif_block (Block): mmcif Block from gemmi.
        label (str): name of the category

    Returns:
        str: Possible error encountered
    """
    if label not in cif_block.get_mmcif_category_names():
        return f"Namespace {label} does not exist."



# software from PDBe: Protein Data Bank in Europe; http://pdbe.org
#
# Copyright 2017 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.
import glob
import os


def tests_dir():
    """
    directory of the tests directory where this python file is.

     Returns:
         str the name of the directory is

    """
    return os.path.dirname(os.path.abspath(__file__))


test_cif = os.path.join(tests_dir(), "ccd_mmcif_test_files", "random_sample")
test_depiction = os.path.join(tests_dir(), "ccd_mmcif_test_files", "depiction_test")
test_cut_down_components_cif = os.path.join(
    tests_dir(), "components_cif", "cut_down_components.cif"
)
test_boundmolecules = os.path.join(
    tests_dir(), "ccd_mmcif_test_files", "boundMolecule_sample"
)


def cif_filename(code):
    path = os.path.join(test_cif, f"{code}.cif")

    return path if os.path.isfile(path) else os.path.join(test_depiction, f"{code}.cif")


def unl_model():
    return os.path.join(tests_dir(), "ccd_mmcif_test_files", "UNL.cif")


def supply_list_of_sample_cifs():
    """
    returns the list of sample pdb ccd cifs for test.

    Args:

    Returns:
        list of filenames
    """
    return sorted(glob.glob(os.path.join(test_cif, "*.cif")))


def supply_list_of_cifs_with_boundmolecules():
    """
    returns a list of cif files with boundmolecules

    Args:

    Returns:
        list of filenames
    """

    return sorted(glob.glob(os.path.join(test_boundmolecules, "*.cif")))



"""Package configuration with the paths to the templates
"""
import os
import pdbeccdutils

general_templates = os.path.join(
    os.path.dirname(pdbeccdutils.__file__), "data", "general_templates"
)
coordgen_templates = os.path.join(
    os.path.dirname(pdbeccdutils.__file__), "data", "coordgen_templates"
)
fragment_library = os.path.join(
    os.path.dirname(pdbeccdutils.__file__), "data", "fragment_library.tsv"
)



import pytest

from pdbeccdutils.core import ccd_reader
from pdbeccdutils.tests.tst_utilities import cif_filename

test_inputs = {
    "ATP": {
        "logp": -2.438,
        "heavy_atom_count": 31,
        "numH_acceptors": 18,
        "numH_donors": 7,
        "num_rotable_bonds": 15,
        "rings_count": 3,
        "TPSA": 279.130,
        "molwt": 506.996,
    },
    "NAG": {
        "logp": -3.078,
        "heavy_atom_count": 15,
        "numH_acceptors": 6,
        "numH_donors": 5,
        "num_rotable_bonds": 7,
        "rings_count": 1,
        "TPSA": 119.250,
        "molwt": 221.09,
    },
}


class TestPropertyCalculation:
    @staticmethod
    @pytest.mark.parametrize("key", test_inputs)
    def test_valid_properties(key):
        physchem_props = ccd_reader.read_pdb_cif_file(
            cif_filename(key)
        ).component.physchem_properties

        assert test_inputs[key]["logp"] == round(physchem_props["CrippenClogP"], 3)
        assert test_inputs[key]["heavy_atom_count"] == physchem_props["NumHeavyAtoms"]
        assert test_inputs[key]["numH_acceptors"] == physchem_props["NumHBA"]
        assert test_inputs[key]["numH_donors"] == physchem_props["NumHBD"]
        assert (
            test_inputs[key]["num_rotable_bonds"] == physchem_props["NumRotatableBonds"]
        )
        assert test_inputs[key]["rings_count"] == physchem_props["NumRings"]
        assert test_inputs[key]["TPSA"] == round(physchem_props["tpsa"], 3)
        assert test_inputs[key]["molwt"] == round(physchem_props["exactmw"], 3)

    @staticmethod
    @pytest.mark.parametrize("key", ["10R", "08T"])
    def test_invalid_properties(key):
        physchem_props = ccd_reader.read_pdb_cif_file(
            cif_filename(key)
        ).component.physchem_properties

        assert physchem_props == {}



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""module handling UniChem mapping
"""

import requests

# these are the resources we agreed on providing as part of the PDBeChem
agreed_resources = {
    "1": ("ChEMBL", "ChEMBL"),
    "2": ("DrugBank", "DrugBank"),
    "4": ("Guide to Pharmacology", "Guide to Pharmacology"),
    "6": ("KEGG LIGAND", "KEGG (Kyoto Encyclopedia of Genes and Genomes) Ligand"),
    "7": ("ChEBI", "ChEBI (Chemical Entities of Biological Interest)."),
    "9": ("ZINC", "ZINC"),
    "22": ("PubChem", "PubChem Compounds"),
    "24": ("NMRShiftDB", "NMRShiftDB"),
    "31": ("BindingDb", "BindingDb"),
    "33": ("LipidMaps", "LipidMaps"),
    "36": ("MetaboLights", "MetaboLights"),
    "37": ("BRENDA", "BRENDA"),
    "38": ("Rhea", "Rhea"),
}

# list taken from https://www.ebi.ac.uk/unichem/ucquery/listSources
all_resource = {
    "1": ("ChEMBL", "ChEMBL"),
    "2": ("DrugBank", "DrugBank"),
    "3": ("PDBe", "Protein Data Bank in Europe"),
    "4": ("Guide to Pharmacology", "Guide to Pharmacology"),
    "5": ("PubChem DOTF", 'PubChem ("Drugs of the Future" subset'),
    "6": ("KEGG LIGAND", "KEGG (Kyoto Encyclopedia of Genes and Genomes) Ligand"),
    "7": ("ChEBI", "ChEBI (Chemical Entities of Biological Interest)."),
    "8": ("NIH", "NIH Clinical Collection"),
    "9": ("ZINC", "ZINC"),
    "10": ("eMolecules", "eMolecules"),
    "11": (
        "IBM",
        "IBM strategic IP insight platform and the National Institutes of Health",
    ),
    "12": ("atlas", "Gene Expression Atlas"),
    "14": ("fdasrs", "FDA/USP Substance Registration System (SRS)"),
    "15": ("SureChEMBL", "SureChEMBL"),
    "17": ("PharmGKB", "PharmGKB"),
    "18": ("HMDB", "Human Metabolome Database (HMDB)"),
    "20": ("Selleck", "Selleck"),
    "21": ("PubChem TPHARMA", 'PubChem ("Thomson Pharma" subset)'),
    "22": ("PubChem", "PubChem Compounds"),
    "23": ("Mcule", "Mcule"),
    "24": ("NMRShiftDB", "NMRShiftDB"),
    "25": ("LINCS", "Library of Integrated Network-based Cellular Signatures"),
    "26": ("ACTor", "ACTor"),
    "27": ("Recon", "Recon"),
    "28": ("MolPort", "MolPort"),
    "29": ("Nikkaji", "Nikkaji"),
    "31": ("BindingDb", "BindingDb"),
    "32": (
        "EPA CompTox Dashboard",
        "EPA (Environmental Protection Agency) CompTox Dashboard",
    ),
    "33": ("LipidMaps", "LipidMaps"),
    "34": ("DrugCentral", "DrugCentral"),
    "35": ("Carotenoid Database", "Carotenoid Database"),
    "36": ("MetaboLights", "MetaboLights"),
    "37": ("BRENDA", "BRENDA"),
    "38": ("Rhea", "Rhea"),
    "39": ("ChemicalBook", "ChemicalBook"),
    "41": ("SwissLipids", "SwissLipids"),
}

url_prefix = "https://www.ebi.ac.uk/unichem/rest/inchikey"


def get_agreed_unichem_mapping(inchikey):
    """Get the agreed mappings for PDBeChem:
        * ChEMBL, DrugBank, KEGG LIGAND, ChEBI, ZINC, N¬MRShiftDB, BindingDB
          MetaboLights, BRENDA, Rhea

    Args:
        inchikey (str): InchiKey (preferably from CIF CCD)

    Returns:
        dict of str: str: Resource mapping. Key is a resource name,
            value is the internal identifier.
    """

    return _download_unichem_mapping(inchikey, agreed_resources)


def get_all_unichem_mapping(inchikey):
    """Get all mappings for PDBeChem

    Args:
        inchikey (str): InchiKey (preferably from CIF CCD)

    Returns:
        dict of str: str: Resource mapping. Key is a resource name,
            value is the internal identifier.
    """

    return _download_unichem_mapping(inchikey, all_resource)


def _download_unichem_mapping(inchikey, resources):
    """Internal function to query unichem server.

    Args:
        inchikey (str): InchiKey (preferably from CIF CCD)
        matched_dict (dict of str: str): Internal UniChem mapping of
            external servers.

    Returns:
        dict of str: [str]: Resource mapping. Key is a resource name,
            value is the list of internal identifiers.
    """
    mapping = []
    url = f"{url_prefix}/{inchikey}"

    try:
        headers = {"Content-Type": "application/json"}
        r = requests.get(url, headers)
        jsonFile = r.json()

        for entity in jsonFile:
            if entity["src_id"] not in resources:
                continue

            resource = resources[entity["src_id"]]
            key = resource[0]
            mapping.append((key, entity["src_compound_id"]))

        return mapping

    except Exception:
        return mapping



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

import csv
import os
from typing import Dict

import rdkit
from rdkit import Chem
from rdkit.Chem import AllChem, rdCoordGen

import pdbeccdutils.utils.config as config
from pdbeccdutils.core.models import FragmentEntry


class FragmentLibrary:
    """Implementation of fragment library."""

    def __init__(
        self,
        path: str = config.fragment_library,
        header: bool = True,
        delimiter: str = "\t",
        quotechar: str = '"',
    ) -> None:
        self.library: Dict[str, FragmentEntry] = {}
        self.name = os.path.basename(path).split(".")[0]
        self._read_in_library(path, header, delimiter, quotechar)

    def _read_in_library(self, path, header, delimiter, quotechar):
        """Fragment library parser

        Args:
            path (str): Path to the fragment library.
            header (bool): Whether or not the library contains a header.
            delimiter (str): Delimiter symbol.
            quotechar (str): Quotechar symbol.
        """
        rdkit.rdBase.DisableLog("rdApp.*")
        rdCoordGen.SetDefaultTemplateFileDir(config.coordgen_templates)
        rdCoordGen.CoordGenParams.coordgenScaling = 33  # to get single bond length 1.5

        with open(path, "r") as csvfile:
            library_reader = csv.reader(
                csvfile, delimiter=delimiter, quotechar=quotechar
            )
            if header:
                next(library_reader)

            for row in library_reader:
                mol = None
                if row[1] == "SMARTS":
                    mol = Chem.MolFromSmarts(row[2])
                else:
                    mol = Chem.MolFromSmiles(row[2])
                    if row[1] == "LIKE":
                        for bond in mol.GetBonds():
                            bond.SetBondType(Chem.rdchem.BondType.UNSPECIFIED)

                Chem.SanitizeMol(mol, catchErrors=True)
                rdCoordGen.AddCoords(mol)

                self.library[row[0]] = FragmentEntry(row[0], row[6], mol)

        rdkit.rdBase.EnableLog("rdApp.*")

    def to_image(self, path, source=""):
        """Export image with all fragments.

        Args:
            path (str): Destination of the image
            source (str): Select a source which fragments are going to
                be drawn.

        """
        use_svg = path[-3:].lower() == "svg"
        if source:
            temp = {k: v for k, v in self.library.items() if v.source == source}
        else:
            temp = self.library

        mols = [v.mol for v in temp.values()]
        names = [v.name for v in temp.values()]
        img = Chem.Draw.MolsToGridImage(
            mols, legends=names, molsPerRow=10, useSVG=use_svg
        )

        if use_svg:
            with open(path, "w") as f:
                f.write(img)
        else:
            img.save(path)

    def generate_conformers(self):
        """Generate 3D coordinates for the fragment library."""
        rdkit.rdBase.DisableLog("rdApp.*")
        options = AllChem.ETKDGv2()
        options.clearConfs = False

        for v in self.library.values():
            try:
                AllChem.EmbedMolecule(v.mol, options)
                AllChem.UFFOptimizeMolecule(v.mol)
            except Exception:
                pass  # don't care if it fails
        rdkit.rdBase.EnableLog("rdApp.*")



"""Test of PARITY method
"""

import pytest

from pdbeccdutils.computations.parity_method import compare_molecules
from pdbeccdutils.core import ccd_reader
from pdbeccdutils.tests.tst_utilities import cif_filename


class TestParity:
    @staticmethod
    @pytest.mark.parametrize(
        "id_1,id_2,score",
        [
            ("MAN", "NAG", 0.74),
            ("NAD", "NAG", 0.09),
            ("000", "000", 1.00),
            ("SAC", "GOL", 0.50),
            ("MAN", "GLC", 1.00),
        ],
    )
    def test_parity_method(id_1, id_2, score):
        c1 = ccd_reader.read_pdb_cif_file(cif_filename(id_1)).component
        c2 = ccd_reader.read_pdb_cif_file(cif_filename(id_2)).component

        parity = compare_molecules(c1.mol, c2.mol)

        assert round(parity.similarity_score, 2) == score



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2019 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import logging
import re
from datetime import date
from typing import Any, Dict, List, Tuple

import rdkit
import gemmi
from rdkit.Chem import BRICS, Draw
from rdkit.Chem.rdMolDescriptors import Properties
from rdkit.Chem.Scaffolds import MurckoScaffold

from pdbeccdutils.core.depictions import DepictionManager, DepictionResult
from pdbeccdutils.core.exceptions import CCDUtilsError
from pdbeccdutils.helpers import mol_tools
from pdbeccdutils.core.fragment_library import FragmentLibrary
from pdbeccdutils.core.models import (
    CCDProperties,
    ConformerType,
    Descriptor,
    ReleaseStatus,
    ScaffoldingMethod,
    SubstructureMapping,
)
from pdbeccdutils.helpers import conversions, drawing
from pdbeccdutils.utils import web_services


class Component:
    """
    Wrapper for the rdkit.Chem.Mol object enabling some of its
    functionality and handling possible erroneous situations.

    Returns:
        Component: instance object
    """

    def __init__(
        self,
        mol: rdkit.Chem.rdchem.Mol,
        ccd_cif_block: gemmi.cif.Block,
        properties: CCDProperties = None,
        descriptors: List[Descriptor] = None,
    ) -> None:
        self.mol = mol
        self._mol_no_h = None
        self.mol2D = None
        self.ccd_cif_block = ccd_cif_block
        self._fragments: Dict[str, SubstructureMapping] = {}
        self._scaffolds: Dict[str, SubstructureMapping] = {}
        self._descriptors: List[Descriptor] = []
        self._inchi_from_rdkit = ""
        self._inchikey_from_rdkit = ""
        self._physchem_properties: Dict[str, Any] = {}
        self._external_mapping: List[Tuple[str, str]] = []

        if descriptors is not None:
            self._descriptors = descriptors

        if properties is not None:
            self._cif_properties = properties

    # region properties
    @property
    def id(self) -> str:
        """Supply the unique identifier for the PDB-CCD,
        for example 'ATP'.
        Obtained from CCD's _chem_comp.id:

        http://mmcif.wwpdb.org/dictionaries/mmcif_std.dic/Items/_chem_comp.id.html

        If not defined then the empty string '' will be returned.

        Returns:
            str: the _chem_comp.id or ''.
        """
        return self._cif_properties.id

    @property
    def name(self) -> str:
        """Supply the 'full name' of the PDB-CCD, for example 'ETHANOL'.
        Obtained from PDB-CCD's _chem_comp.name:

        http://mmcif.wwpdb.org/dictionaries/mmcif_std.dic/Items/_chem_comp.name.html

        If not defined then the empty string '' will be returned.

        Returns:
            str: the _chem_comp.name or ''.
        """
        return self._cif_properties.name

    @property
    def formula(self) -> str:
        """Supply the chemical formula for the PDB-CCD,
        for example 'C2 H6 O'.
        Obtained from PDB-CCD's _chem_comp.formula:

        http://mmcif.wwpdb.org/dictionaries/mmcif_std.dic/Items/_chem_comp.formula.html

        If not defined then the empty string '' will be returned.

        Returns:
            str: the _chem_comp.formula or ''.
        """
        return self._cif_properties.formula

    @property
    def pdbx_release_status(self) -> ReleaseStatus:
        """Supply the pdbx_release_status for the PDB-CCD.
        Obtained from PDB-CCD's _chem_comp.pdbx_rel_status:

        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx.dic/Items/_chem_comp.pdbx_release_status.html

        Returns:
            pdbeccdutils.core.enums.ReleaseStatus: enum of the release
            status (this includes NOT_SET if no value is defined).
        """
        return self._cif_properties.pdbx_release_status

    @property
    def modified_date(self) -> date:
        """Supply the _pdbx_chem_comp_descriptor category for the PDB-CCD
        Obtained from PDB-CCD's _pdbx_chem_comp_descriptor:

        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_chem_comp.pdbx_modified_date.html

        Returns:
            datetime.date: Date of the last entrie's modification.
        """
        return self._cif_properties.modified_date

    @property
    def descriptors(self) -> List[Descriptor]:
        """Supply the pdbx_modified_date for the PDB-CCD
        Obtained from PDB-CCD's _chem_comp.pdbx_modified_date:

        http://mmcif.rcsb.org/dictionaries/mmcif_pdbx.dic/Items/_pdbx_chem_comp_descriptor.program_version.html

        Returns:
            list[Descriptor]: List of descriptors for a given entry.
        """
        return self._descriptors

    @property
    def inchikey(self) -> str:
        """Supply the InChIKey for the PDB-CCD.
        Obtained from `PDB-CCD's _pdbx_chem_comp_descriptor` table line
        with `_pdbx_chem_comp_descriptor.type=InChIKey`, see:

        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx.dic/Items/_pdbx_chem_comp_descriptor.type.html

        If not defined then the empty string '' will be returned.

        Returns:
            str: the InChIKey or ''.
        """
        return next((x.value for x in self._descriptors if x.type == "InChIKey"), "")

    @property
    def inchi(self) -> str:
        """Supply the InChI for the PDB-CCD.
        Obtained from `PDB-CCD's _pdbx_chem_comp_descriptor` table line
        with `_pdbx_chem_comp_descriptor.type=InChI`, see:

        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx.dic/Items/_pdbx_chem_comp_descriptor.type.html

        If not defined then the empty string '' will be returned.

        Returns:
            str: the InChI or ''.
        """
        return next((x.value for x in self._descriptors if x.type == "InChI"), "")

    @property
    def inchi_from_rdkit(self) -> str:
        """Provides the InChI worked out by RDKit.

        Returns:
            str: the InChI or empty '' if there was an error finding it.
        """
        if not self._inchi_from_rdkit:
            try:
                self._inchi_from_rdkit = rdkit.Chem.inchi.MolToInchi(self.mol)
            except ValueError:
                self._inchi_from_rdkit = ""
        return self._inchi_from_rdkit

    @property
    def inchikey_from_rdkit(self) -> str:
        """Provides the InChIKey worked out by RDKit.

        Returns:
            str: the InChIKey or '' if there was an error finding it.
        """
        if not self._inchikey_from_rdkit:
            inchi = self.inchi_from_rdkit
            if inchi != "ERROR":
                self._inchikey_from_rdkit = rdkit.Chem.inchi.InchiToInchiKey(inchi)
            else:
                self._inchikey_from_rdkit = ""
            if self._inchikey_from_rdkit is None:
                self._inchikey_from_rdkit = ""
        return self._inchikey_from_rdkit

    @property
    def released(self) -> bool:
        """Tests pdbx_release_status is REL.

        Returns:
            bool: True if PDB-CCD has been released.
        """
        return self._cif_properties.pdbx_release_status == ReleaseStatus.REL

    @property
    def mol_no_h(self) -> rdkit.Chem.rdchem.Mol:
        """RDKit mol object without hydrogens

        Returns:
            rdkit.Chem.rdchem.Mol: RDKit mol object with stripped Hs.
        """
        if self._mol_no_h is None:
            no_h = rdkit.Chem.RemoveHs(self.mol, sanitize=False)
            rdkit.Chem.SanitizeMol(no_h, catchErrors=True)
            self._mol_no_h = no_h

        return self._mol_no_h

    @property
    def number_atoms(self) -> int:
        """Supplies the number of atoms in the _chem_comp_atom table

        Returns:
            int: the number of atoms in the PDB-CCD
        """
        return self.mol.GetNumAtoms()

    @property
    def fragments(self) -> List[SubstructureMapping]:
        """Lists matched fragments and atom names.

        Returns:
            list[SubstructureMapping]: Substructure mapping for
            all discovered fragments.
        """
        return self._id_to_name_mapping(self._fragments)

    @property
    def scaffolds(self) -> List[SubstructureMapping]:
        """Lists matched scaffolds and atom names

        Returns:
            list[SubstructureMapping]: List of substructure mappings.
        """
        return self._id_to_name_mapping(self._scaffolds)

    @property
    def atoms_ids(self) -> Tuple[Any, ...]:
        """Supplies a list of the atom_ids obtained from
        `_chem_comp_atom.atom_id`, see:

        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx.dic/Categories/chem_comp_atom.html

        The order will reflect the order in the input PDB-CCD.

        The atom_id is also also know as 'atom_name', standard amino
        acids have main chain atom names 'N CA C O'

        Returns:
            tuple[str]: `atom_id's` for the PDB-CCD
        """
        return tuple(atom.GetProp("name") for atom in self.mol.GetAtoms())

    @property
    def physchem_properties(self):
        """RDKit calculated properties related to the CCD compound

        Returns:
            dict[str, float]: A list of RDKit calculated properties
        """
        if not self._physchem_properties:
            try:
                properties = Properties()
                self._physchem_properties = dict(
                    zip(
                        properties.GetPropertyNames(),
                        properties.ComputeProperties(self.mol),
                    )
                )
                self._physchem_properties["NumHeavyAtoms"] = float(
                    self.mol.GetNumHeavyAtoms()
                )
            except (RuntimeError, ValueError):
                return {}

        return self._physchem_properties

    @property
    def external_mappings(self):
        """List external mappings provided by UniChem. fetch_external_mappings()
        was not called before only agreed mapping is retrieved.

        Returns:
            list[tuple[str]]: UniChem mappings
        """

        return self._external_mapping

    @external_mappings.setter
    def external_mappings(self, value):
        """Set mapping for this component obtained with a different mean
        but internal use of UniChem.

        Args:
            list[tuple[str]]: UniChem mappings
        """
        self._external_mapping = value

    # endregion properties

    def fetch_external_mappings(self, all_mappings=False):
        """Retrieve external mapping through UniChem based on the InChi Key.

        Args:
            all_mappings (bool, optional): Get UniChem mappings. Defaults to False.

        Returns:
            dict[str, str]: Return resource ids pairing established by UniChem.
        """
        if all_mappings:
            self._external_mapping = web_services.get_all_unichem_mapping(self.inchikey)
        else:
            self._external_mapping = web_services.get_agreed_unichem_mapping(
                self.inchikey
            )

        return self._external_mapping

    def inchikey_from_rdkit_matches_ccd(self, connectivity_only: bool = False) -> bool:
        """Checks whether inchikey matches between ccd and rdkit

        Args:
            connectivity_only (bool): restrict to the first 14 character - the connectivity information.

        Returns:
            bool: True for match
        """
        if self.inchikey is None or self.inchikey_from_rdkit == "ERROR":
            return False
        if connectivity_only:
            if len(self.inchikey) < 14 or len(self.inchikey_from_rdkit) < 14:
                return False
            if self.inchikey[:14] != self.inchikey_from_rdkit[:14]:
                return False

        if self.inchikey != self.inchikey_from_rdkit:
            return False

        return True

    def compute_2d(
        self, manager: DepictionManager, remove_hs: bool = True
    ) -> DepictionResult:
        """Compute 2d depiction of the component using DepictionManager
        instance.

        Args:
            manager (DepictionManager): Instance of the ligand depiction
                class.
            remove_hs (bool, optional): Defaults to True. Remove
                hydrogens prior to depiction.

        Returns:
            DepictionResult: Object with the details about depiction process.
        """
        mol_copy = rdkit.Chem.RWMol(self.mol)
        if remove_hs:
            mol_copy = rdkit.Chem.RemoveHs(
                mol_copy, updateExplicitCount=True, sanitize=False
            )
            rdkit.Chem.SanitizeMol(mol_copy, catchErrors=True)

        result_log = manager.depict_molecule(self.id, mol_copy)

        self.mol2D = result_log.mol

        return result_log

    def export_2d_svg(
        self,
        file_name: str,
        width: int = 500,
        names: bool = False,
        wedge_bonds: bool = True,
        atom_highlight: Dict[Any, Tuple] = None,
        bond_highlight: Dict[Tuple, Tuple] = None,
    ):
        """Save 2D depiction of the component as an SVG file. Component
        id is generated in case the image cannot be drawn.

        Args:
            file_name (str): path to store 2d depiction
            width (int, optional): Defaults to 500. Width of a frame in pixels.
            names (bool, optional): Defaults to False. Whether or not to
                include atom names in depiction. If atom name is not set, element symbol is used instead.
            wedge_bonds (bool, optional): Defaults to True. Whether or not
                the molecule should be depicted with bond wedging.
            atomHighlight (:obj:`dict` of :obj:`tuple` of :obj:`float`, optional):
                Defaults to None. Atoms names to be highlighted along
                with colors in RGB. e.g. {'CA': (0.5, 0.5, 0.5)} or {0: (0.5, 0.5, 0.5)}
            bondHighlight (:obj:`dict` of :obj:`tuple` of :obj:`float`, optional):
                Defaults to None. Bonds to be highlighted along with
                colors in RGB. e.g. {('CA', 'CB'): (0.5, 0.5, 0.5)} or {(0, 1): (0.5, 0.5, 0.5)}

        Raises:
            CCDUtilsError: If bond or atom does not exist.
        """
        get_atom_name = (
            lambda a: a.GetProp("name")
            if a.HasProp("name")
            else a.GetSymbol() + str(a.GetIdx())
        )

        if self.mol2D is None:
            drawing.save_no_image(file_name, self.id, width)
            return

        drawer = Draw.rdMolDraw2D.MolDraw2DSVG(width, width)
        options = drawer.drawOptions()
        atom_mapping = {
            get_atom_name(a): i for i, a in enumerate(self.mol2D.GetAtoms())
        }

        atom_highlight = {} if atom_highlight is None else atom_highlight
        bond_highlight = {} if bond_highlight is None else bond_highlight

        if width < 201:
            options.bondLineWidth = 1

        if all(isinstance(i, str) for i in atom_highlight.keys()):
            atom_highlight = {atom_mapping[k]: v for k, v in atom_highlight.items()}
        else:
            atom_highlight = {}

        if bond_highlight:
            if all(
                isinstance(i[0], str) and isinstance(i[1], str)
                for i in bond_highlight.keys()
            ):
                temp_highlight = {}
                for k, v in bond_highlight.items():
                    bond = self.mol2D.GetBondBetweenAtoms(
                        atom_mapping[k[0]], atom_mapping[k[1]]
                    )
                    if bond is None:
                        raise CCDUtilsError(
                            "Bond between {} and {} does not exist".format(k[0], k[1])
                        )
                    temp_highlight[bond.GetIdx()] = v
                bond_highlight = temp_highlight

        if names:
            for i, a in enumerate(self.mol2D.GetAtoms()):
                atom_name = get_atom_name(a)
                options.atomLabels[i] = atom_name
                a.SetProp("molFileAlias", atom_name)

        drawing.draw_molecule(
            self.mol2D, drawer, file_name, wedge_bonds, atom_highlight, bond_highlight
        )

    def export_2d_annotation(self, file_name: str, wedge_bonds: bool = True) -> None:
        """Generates 2D depiction in JSON format with annotation of
        bonds and atoms to be redrawn in the interactions component.

        Args:
            file_name (str): Path to the file
        """
        w, h = drawing.get_drawing_scale(self.mol2D)
        drawer = Draw.MolDraw2DSVG(w, h)
        drawer.drawOptions().includeAtomTags = True
        try:
            tmp = rdkit.Chem.Draw.PrepareMolForDrawing(
                self.mol2D, wedgeBonds=wedge_bonds, kekulize=True, addChiralHs=False
            )
        except (RuntimeError, ValueError):
            tmp = rdkit.Chem.Draw.PrepareMolForDrawing(
                self.mol2D, wedgeBonds=False, kekulize=True, addChiralHs=False
            )
        drawer.DrawMolecule(tmp)
        drawer.TagAtoms(self.mol2D)
        drawer.FinishDrawing()
        svg = drawer.GetDrawingText()
        json_repr = drawing.convert_svg(svg, self.id, self.mol2D)

        with open(file_name, "w") as fp:
            json.dump(json_repr, fp, indent=4, sort_keys=True)

    def compute_3d(self, version="v3") -> bool:
        """Generate 3D coordinates using EKTDG method. Version can be specified

        Args:
            version (str, optional): Version of EKTDG to be used.
                Defaults to "v3".

        Returns:
            bool: Result of the structure generation process.
        """

        if version == "v3":
            options = rdkit.Chem.AllChem.ETKDGv3()
        elif version == "v2":
            options = rdkit.Chem.AllChem.ETKDGv2()
        else:
            options = rdkit.Chem.AllChem.ETKDGv2()

        options.clearConfs = False
        conf_id = -1

        try:
            conf_id = rdkit.Chem.AllChem.EmbedMolecule(self.mol, options)
            rdkit.Chem.AllChem.UFFOptimizeMolecule(
                self.mol, confId=conf_id, maxIters=1000
            )

        except RuntimeError:
            pass  # Force field issue here
        except ValueError:
            pass  # sanitization issue here

        if conf_id != -1:
            conformer = self.mol.GetConformer(conf_id)
            conformer.SetProp("name", ConformerType.Computed.name)
            conformer.SetProp("coord_generation", f"ETKDG{version}")

            return True

        return False

    def has_degenerated_conformer(self, c_type: ConformerType) -> bool:
        """
        Determine if given conformer has missing coordinates or is
        missing completelly from the rdkit.Mol object. This can
        be used to determine, whether or not the coordinates should be
        regenerated.

        Args:
            type (ConformerType): type of conformer
                to be inspected.

        Returns:
            bool: True if more then 1 atom has coordinates [0, 0, 0]
        """
        ok_conformer = False

        for c in self.mol.GetConformers():
            if c.GetProp("name") == c_type.name:
                ok_conformer = mol_tools.is_degenerate_conformer(c)
                break

        return ok_conformer

    def get_conformer(self, c_type) -> rdkit.Chem.rdchem.Conformer:
        """
        Retrieve an rdkit object for a deemed conformer.

        Args:
            c_type (ConformerType): Conformer type to be retrieved.

        Raises:
            ValueError: If conformer does not exist

        Returns:
            rdkit.Chem.rdchem.Conformer: RDKit conformer object
        """
        for c in self.mol.GetConformers():
            try:
                if c.GetProp("name") == c_type.name:
                    return c
            except KeyError:
                pass

        raise ValueError(f"Conformer {c_type.name} does not exist.")

    def locate_fragment(
        self, mol: rdkit.Chem.rdchem.Mol
    ) -> List[List[rdkit.Chem.rdchem.Atom]]:
        """
        Identify substructure match in the component.

        Args:
            mol (rdkit.Chem.rdchem.Mol): Fragment to be matched with
                structure

        Returns:
            list[list[rdkit.Chem.rdchem.Atom]]: List of fragments
            identified in the component as a list of atoms.
        """
        result = []
        if mol is None:
            return []

        matches = self.mol.GetSubstructMatches(mol)

        for m in matches:
            result.append([self.mol.GetAtomWithIdx(idx) for idx in m])

        return result

    def library_search(
        self, fragment_library: FragmentLibrary
    ) -> List[SubstructureMapping]:
        """Identify fragments from the fragment library in this component

        Args:
            fragment_library (FragmentLibrary): Fragment library.

        Returns:
            list[SubstructureMapping]: Matches found in this run
        """
        temp = {}
        for v in fragment_library.library.values():
            try:
                matches = self.mol_no_h.GetSubstructMatches(v.mol)

                if not matches:
                    continue

                key = f"{fragment_library.name}_{v.name}"
                if key not in self._fragments:
                    temp[key] = SubstructureMapping(
                        v.name, rdkit.Chem.MolToSmiles(v.mol), v.source, matches
                    )

            except Exception:
                logging.warning(f"Error mapping fragment {v.name}.")

        self._fragments.update(temp)

        return list(temp.values())

    def get_scaffolds(self, scaffolding_method=ScaffoldingMethod.MurckoScaffold):
        """Compute deemed scaffolds for a given compound.

        Args:
            scaffolding_method (ScaffoldingMethod, optional):
                Defaults to MurckoScaffold. Scaffolding method to use

        Returns:
            list[rdkit.Chem.rdchem.Mol]: Scaffolds found in the component.
        """
        try:
            scaffolds = []

            if scaffolding_method == ScaffoldingMethod.MurckoScaffold:
                scaffolds = [(MurckoScaffold.GetScaffoldForMol(self.mol_no_h))]

            elif scaffolding_method == ScaffoldingMethod.MurckoGeneric:
                scaffolds = [(MurckoScaffold.MakeScaffoldGeneric(self.mol_no_h))]

            elif scaffolding_method == ScaffoldingMethod.Brics:
                scaffolds = BRICS.BRICSDecompose(self.mol_no_h)
                brics_smiles = [
                    re.sub(r"(\[[0-9]*\*\])", "[H]", i) for i in scaffolds
                ]  # replace dummy atoms with H's to get matches https://sourceforge.net/p/rdkit/mailman/message/35261974/
                brics_mols = [rdkit.Chem.MolFromSmiles(x) for x in brics_smiles]

                for mol in brics_mols:
                    rdkit.Chem.RemoveHs(mol)

                brics_hits = [self.mol_no_h.GetSubstructMatches(i) for i in brics_mols]

                for index, brics_hit in enumerate(brics_hits):
                    smiles = rdkit.Chem.MolToSmiles(brics_mols[index])
                    name = scaffolding_method.name
                    source = "RDKit scaffolds"
                    key = f"{name}_{smiles}"
                    brics_hit = conversions.listit(brics_hit)

                    if not smiles:
                        continue

                    if key not in self._scaffolds:
                        self._scaffolds[key] = SubstructureMapping(
                            name, smiles, source, brics_hit
                        )

                return brics_mols

            for s in scaffolds:
                scaffold_atom_names = [atom.GetProp("name") for atom in s.GetAtoms()]
                mapping = []
                for at_name in scaffold_atom_names:
                    idx = [
                        atom.GetIdx()
                        for atom in self.mol.GetAtoms()
                        if atom.GetProp("name") == at_name
                    ][0]
                    mapping.append(idx)

                smiles = rdkit.Chem.MolToSmiles(s)
                name = scaffolding_method.name
                source = "RDKit scaffolds"

                if not smiles:
                    continue

                if name in self._scaffolds:
                    self._scaffolds[name].mappings.append(mapping)
                else:
                    self._scaffolds[name] = SubstructureMapping(
                        name, smiles, source, [mapping]
                    )

            return scaffolds

        except (RuntimeError, ValueError):
            raise CCDUtilsError(
                f"Computing scaffolds using method {scaffolding_method.name} failed."
            )

    def _id_to_name_mapping(self, struct_mapping):
        """Lists matched scaffolds and atom names

        Args:
            struct_mapping (dict[str, SubstructureMapping]): Basic mapping
                for fragments/scaffolds.

        Returns:
            dict[str, SubstructureMapping]: Dictionary with scaffold names
            and matched atoms.
        """
        res = []

        for v in struct_mapping.values():
            mappings = []

            for m in v.mappings:
                atom_names = [self.mol.GetAtomWithIdx(idx).GetProp("name") for idx in m]
                mappings.append(atom_names)
            res.append(SubstructureMapping(v.name, v.smiles, v.source, mappings))

        return res



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2021 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Set of methods for molecular sanitization and work with conformers
"""

import re
import sys
from io import StringIO

import numpy as np
import rdkit

empty_coords = rdkit.Chem.rdGeometry.Point3D(0, 0, 0)
METALS_SMART = (
    "[Li,Na,K,Rb,Cs,F,Be,Mg,Ca,Sr,Ba,Ra,Sc,Ti,V,Cr,Mn,Fe,Co,Ni,Cu,Zn,Al,Ga,Y,Zr,Nb,Mo,"
    "Tc,Ru,Rh,Pd,Ag,Cd,In,Sn,Hf,Ta,W,Re,Os,Ir,Pt,Au,Hg,Tl,Pb,Bi]"
)


def is_degenerate_conformer(conformer):
    """
    Determine if given conformer has missing coordinates or is
    missing completelly from the rdkit.Mol object. This can
    be used to determine, whether or not the coordinates should be
    regenerated.

    Args:
        conformer (rdkit.Chem.rdchem.Conformer): conformer to check

    Returns:
        bool: true if more then 1 atom has coordinates [0, 0, 0]
    """
    counter = 0

    try:
        for i in range(conformer.GetNumAtoms()):
            pos = conformer.GetAtomPosition(i)
            if pos.Distance(empty_coords) == 0.0:
                counter += 1

            if counter > 1:
                return True
    except ValueError:  # Conformer does not exist
        return False

    return False


def sanitize(rwmol):
    """
    Attempts to sanitize mol in place. RDKit's standard error can be
    processed in order to find out what went wrong with sanitization
    to fix the molecule.

    Args:
        rwmol (rdkit.Chem.rdchem.RWMol): rdkit molecule to be sanitized

    Returns:
        bool: Result of the sanitization process.
    """
    success = False

    try:
        success = fix_molecule(rwmol)

        if not success:
            return False

        rdkit.Chem.Kekulize(rwmol)
        # rdkit.Chem.rdmolops.AssignAtomChiralTagsFromStructure(rwmol, confId=0)

        # find correct conformer to assign stereochemistry
        # ideal conformer comes first
        conformers = rwmol.GetConformers()
        conformer_id = -1

        if is_degenerate_conformer(conformers[0]):
            conformer_id = conformers[1].GetId()
        else:
            conformer_id = conformers[0].GetId()

        rdkit.Chem.rdmolops.AssignStereochemistryFrom3D(rwmol, conformer_id)

    except Exception as e:
        print(e, file=sys.stderr)
        return False

    return success


def fix_molecule(rwmol: rdkit.Chem.rdchem.RWMol):
    """
    Single molecule sanitization process. Presently, only valence
    errors are taken care are of.

    Args:
        rwmol (rdkit.Chem.rdchem.RWMol): rdkit molecule to be
            sanitized

    Returns:
        bool: Whether or not sanitization succeeded
    """
    attempts = 10
    success = False
    saved_std_err = sys.stderr
    log = sys.stderr = StringIO()
    rdkit.rdBase.LogToPythonStderr()

    while (not success) and attempts >= 0:
        sanitization_result = rdkit.Chem.SanitizeMol(rwmol, catchErrors=True)

        if sanitization_result == 0:
            sys.stderr = saved_std_err
            return True

        sanitization_failures = re.findall("[a-zA-Z]{1,2}, \\d+", log.getvalue())
        if not sanitization_failures:
            sys.stderr = saved_std_err
            return False

        for sanitization_failure in sanitization_failures:
            split_object = sanitization_failure.split(",")  # [0] element [1] valency
            element = split_object[0]
            valency = int(split_object[1].strip())

            smarts_metal_check = rdkit.Chem.MolFromSmarts(
                METALS_SMART + "~[{}]".format(element)
            )
            metal_atom_bonds = rwmol.GetSubstructMatches(smarts_metal_check)
            rdkit.Chem.SanitizeMol(
                rwmol, sanitizeOps=rdkit.Chem.SanitizeFlags.SANITIZE_CLEANUP
            )

            for metal_index, atom_index in metal_atom_bonds:
                metal_atom = rwmol.GetAtomWithIdx(metal_index)
                erroneous_atom = rwmol.GetAtomWithIdx(atom_index)

                # change the bond type to dative
                bond = rwmol.GetBondBetweenAtoms(
                    metal_atom.GetIdx(), erroneous_atom.GetIdx()
                )
                bond.SetBondType(rdkit.Chem.BondType.SINGLE)

                if erroneous_atom.GetExplicitValence() == valency:
                    erroneous_atom.SetFormalCharge(erroneous_atom.GetFormalCharge() + 1)
                    metal_atom.SetFormalCharge(metal_atom.GetFormalCharge() - 1)

        attempts -= 1

    sys.stderr = saved_std_err

    return False


def fix_conformer(conformer):
    """In place fixing of rdkit conformer.
    In certain cases the resulting conformer (mainly from depiction process)
    can contain not valid atom coordinatesp [NaN, NaN, NaN]. This
    results in errors in downstream processes so that it is easier
    to fix it when the problem occurs

    Args:
        conformer (rdkit.Chem.rdchem.Conformer): RDKit conformer
    """
    positions = conformer.GetPositions()

    for index, pos in enumerate(positions):
        if all(np.isnan(pos)):
            new_pos = rdkit.Chem.rdGeometry.Point3D(0, 0, 0)
            conformer.SetAtomPosition(index, new_pos)



import rdkit
import numpy as np
from pdbeccdutils.helpers.mol_tools import fix_conformer


class TestRDKitFixtures:
    @staticmethod
    def test_conformer_is_broken_ion():
        mol = rdkit.Chem.RWMol()
        atom = rdkit.Chem.Atom("H")

        mol.AddAtom(atom)

        conformer = rdkit.Chem.Conformer(1)
        atom_position = rdkit.Chem.rdGeometry.Point3D(np.NaN, np.NaN, np.NaN)
        conformer.SetAtomPosition(0, atom_position)
        mol.AddConformer(conformer, assignId=True)

        m = mol.GetMol()

        c = m.GetConformer(0)
        fix_conformer(c)

        assert c.GetAtomPosition(0).x == 0.0
        assert c.GetAtomPosition(0).y == 0.0
        assert c.GetAtomPosition(0).z == 0.0

    @staticmethod
    def test_conformer_has_broken_atom():
        mol = rdkit.Chem.RWMol()
        o = rdkit.Chem.Atom("O")
        h = rdkit.Chem.Atom("H")

        mol.AddAtom(o)
        mol.AddAtom(h)
        mol.AddBond(0, 1, rdkit.Chem.rdchem.BondType(1))

        conformer = rdkit.Chem.Conformer(1)
        o_position = rdkit.Chem.rdGeometry.Point3D(1, 2, 3)
        h_position = rdkit.Chem.rdGeometry.Point3D(np.NaN, np.NaN, np.NaN)

        conformer.SetAtomPosition(0, o_position)
        conformer.SetAtomPosition(1, h_position)
        mol.AddConformer(conformer, assignId=True)

        m = mol.GetMol()

        c = m.GetConformer(0)
        fix_conformer(c)

        assert c.GetAtomPosition(0).x != 0.0
        assert c.GetAtomPosition(0).y != 0.0
        assert c.GetAtomPosition(0).z != 0.0

        assert c.GetAtomPosition(1).x == 0.0
        assert c.GetAtomPosition(1).y == 0.0
        assert c.GetAtomPosition(1).z == 0.0



"""Common fixtures shared among all the tests
"""
import pytest
from pdbeccdutils.core import ccd_reader, bm_reader
from pdbeccdutils.core.fragment_library import FragmentLibrary
from pdbeccdutils.tests.tst_utilities import supply_list_of_sample_cifs
from pdbeccdutils.tests.tst_utilities import supply_list_of_cifs_with_boundmolecules

sample_ccd_cifs = supply_list_of_sample_cifs()
sample_cifs_with_boundmolecules = supply_list_of_cifs_with_boundmolecules()
problematic_ids = ["UNL", "NA", "SY9", "10R", "ASX", "0KA"]


@pytest.fixture(scope="session", params=sample_ccd_cifs)
def component(request):
    reader = ccd_reader.read_pdb_cif_file(request.param)
    c = reader.component

    if c.id not in problematic_ids:
        assert reader.warnings == []

    return c


@pytest.fixture(scope="session", params=sample_cifs_with_boundmolecules)
def boundMolecules(request):
    reader_list = bm_reader.read_pdb_updated_cif_file(request.param)
    bm_components = []
    for reader_result in reader_list:
        c = reader_result.component
        assert reader_result.warnings == []
        bm_components.append(c)

    return bm_components


@pytest.fixture(scope="session")
def library():
    return FragmentLibrary()



"""Small toolkit to update `database` with Pubchem depictions with within
the pdbeccdutils. Based on the inchikey provided in the CCD
a corresponding 2D layout is downloaded from PubChem database.
"""
import argparse
import os
import sys

from pdbeccdutils.utils.pubchem_downloader import PubChemDownloader


def main():
    """Main method of the whole process."""
    parser = argparse.ArgumentParser(
        description="PDBe downloader of pubchem depictions"
    )
    parser.add_argument(
        "--components_dir",
        type=str,
        default="",
        help="Path to the directory with CCD files",
        required=False,
    )
    parser.add_argument(
        "--ccd", type=str, default="", help="Path to the CCD file", required=False
    )
    parser.add_argument(
        "--pubchem_templates",
        type=str,
        help="Path to the pubchem templates.",
        required=True,
    )

    config = parser.parse_args()

    pubchem = PubChemDownloader(config.pubchem_templates)
    if os.path.isdir(config.components_dir):
        pubchem.update_ccd_dir(config.components_dir)
        sys.exit(os.EX_OK)
    elif os.path.isfile(config.ccd):
        pubchem.update_ccd_file(config.ccd)
        sys.exit(os.EX_OK)
    else:
        print(
            "Either components_dir or components.cif file needs to be set as a source of data.",
            file=sys.stderr,
        )
        sys.exit(os.EX_USAGE)


if __name__ == "__main__":
    main()



"""
load EOH.cif from file and test important cif item.
"""
import pytest

from pdbeccdutils.core import ccd_reader, ccd_writer
from pdbeccdutils.core.models import ConformerType, ReleaseStatus
from pdbeccdutils.tests.tst_utilities import cif_filename


@pytest.fixture(scope="module")
def component_eoh():
    """
    loads PDB-CCD for ethanol eoh.cif and returns the
    component object

    Returns:
        pdbeccdutils.core.component.Component: component object for
        eoh.cif
    """
    cif_file = cif_filename("EOH")
    reader = ccd_reader.read_pdb_cif_file(cif_file)
    assert reader.warnings == []
    assert reader.errors == []
    component = reader.component

    return component


def test_chem_comp_dot_id(component_eoh):
    """test retrieval of _chem_comp.id EOH"""
    assert component_eoh.id == "EOH"


chem_comp_items = [
    ("id", "EOH"),
    ("name", "ETHANOL"),
    ("formula", "C2 H6 O"),
    ("pdbx_release_status", ReleaseStatus.REL),
]


@pytest.mark.parametrize("attribute, expected", chem_comp_items)
def test_chem_comp_items(attribute, expected, component_eoh):
    """for EOH.cif tests items from _chem_comp:
    _chem_comp.id                                    EOH
    _chem_comp.name                                  ETHANOL
    _chem_comp.formula                               "C2 H6 O"
    _chem_comp.pdbx_release_status                   REL
    """
    assert hasattr(component_eoh, attribute)
    assert getattr(component_eoh, attribute) == expected


def test_released_is_true(component_eoh):
    assert component_eoh.released is True


def test_inchikey(component_eoh):
    """
    test InChIKey retrieval from _pdbx_chem_comp_descriptor line:
    EOH InChIKey InChI 1.03  LFQSCWFLJHTTHZ-UHFFFAOYSA-N
    """
    assert component_eoh.inchikey == "LFQSCWFLJHTTHZ-UHFFFAOYSA-N"


def test_eoh_has_nine_atoms(component_eoh):
    """test number of atoms in the _chem_comp_atom table"""
    assert component_eoh.number_atoms == 9


def test_eoh_atom_ids(component_eoh):
    assert component_eoh.atoms_ids == (
        "C1",
        "C2",
        "O",
        "H11",
        "H12",
        "H21",
        "H22",
        "H23",
        "HO",
    )


def test_inchikeys_from_rdkit_and_ccd_match(component_eoh):
    assert component_eoh.inchikey == component_eoh.inchikey_from_rdkit


def test_to_sdf_string_ideal_no_h(component_eoh):
    sdf_string = ccd_writer.to_sdf_str(component_eoh)
    # there should be two carbons and no hydrogen atoms:
    assert sdf_string.count(" C ") == 2
    assert sdf_string.count(" O ") == 1
    assert sdf_string.count(" H ") == 0
    # check x and y coordinates of the Oxygen
    assert "1.130" in sdf_string
    assert "0.315" in sdf_string


def test_to_sdf_string_ideal_with_h(component_eoh):
    sdf_string = ccd_writer.to_sdf_str(component_eoh, remove_hs=False)
    # six hydrogen atoms:
    assert sdf_string.count(" H ") == 6
    # check z coordinate of the first hydrogen atom
    assert "0.890" in sdf_string


def test_to_sdf_string_model_no_h(component_eoh):
    sdf_string = ccd_writer.to_sdf_str(component_eoh, conf_type=ConformerType.Model)
    # there should be two carbons and no hydrogen atoms:
    assert sdf_string.count(" C ") == 2
    assert sdf_string.count(" O ") == 1
    assert sdf_string.count(" H ") == 0
    # check x and y coordinates of the Oxygen
    assert "15.861" in sdf_string
    assert "8.256" in sdf_string



"""Lightweight implementation of the `parity method <https://doi.org/10.1016/j.str.2018.02.009>`_ provided by
Jon Tyzack.
"""

from rdkit import Chem
from rdkit.Chem import rdFMCS

from pdbeccdutils.core.models import ParityResult


def _get_matches(mol, smarts):
    """Gets the subgraph match for the parity method.

    Args:
        mol (rdkit.Chem.rdchem.Mol): Molecule to be queried
        smarts (str): Molecule representation in SMARTS.

    Returns:
        int: Molecular subgraph matches.
    """

    patt = Chem.MolFromSmarts(smarts)
    matches = mol.GetSubstructMatches(patt, uniquify=False)
    return matches


def _generate_sim_score(template, query, smarts):
    """Given the two moleculs and their common subgraph represented as
    smarts string returns similarity score.

    Args:
        template (rdkit.Chem.rdchem.Mol): template molecule.
        query (rdkit.Chem.Mol): query molecule.
        smarts (str): common subgraph in the SMARTS format.

    Returns:
        tuple[dict[int, int], float]: atom id mapping for a common
        subgraph and the similarity score.
    """

    if template.GetNumAtoms() == 1:
        smarts = Chem.MolToSmarts(template)
    elif query.GetNumAtoms() == 1:
        smarts = Chem.MolToSmarts(query)
    if smarts is None:
        substructure = {}
        best_sim_score = 0.0
    else:
        matches_1 = _get_matches(template, smarts)
        matches_2 = _get_matches(query, smarts)
        best_matches = 0
        substructure = {}
        size_1 = template.GetNumAtoms()
        size_2 = query.GetNumAtoms()

        for match_1 in matches_1:
            for match_2 in matches_2:
                matches = 0
                for i, value in enumerate(match_1):
                    atom_1 = template.GetAtomWithIdx(value)
                    atom_2 = query.GetAtomWithIdx(match_2[i])
                    symbol_1 = atom_1.GetSymbol()
                    symbol_2 = atom_2.GetSymbol()
                    if symbol_1 == symbol_2:
                        matches += 1
                if matches > best_matches:
                    best_matches = matches
                    substructure = dict(zip(match_1, match_2))
        best_sim_score = float(best_matches) / float(size_1 + size_2 - best_matches)

    return (substructure, best_sim_score)


def compare_molecules(template, query, thresh=0.01, exact_match=False):
    """Given the two molecules calculates their similarity score.
    If expected similarity score is lower than the threshold nothing
    is calculated.

    Args:
        template (rdkit.Chem.rdchem.Mol): Template molecule
        query (rdkit.Chem.rdchem.Mol): Query molecule
        thresh (float, optional): Threshold score for
            the match to be considered. Defaults to 0.01.
        exact_match (bool, optional): Controls whether atom type and
            bond order should be checked too. Defaults to False.

    Returns:
        ParityResult: Result of the PARITY comparison.
    """

    template_atoms = template.GetNumAtoms()
    query_atoms = query.GetNumAtoms()

    min_num_atoms = min(template_atoms, query_atoms)
    max_sim_score = float(min_num_atoms) / float(
        template_atoms + query_atoms - min_num_atoms
    )

    if max_sim_score < thresh:
        return ParityResult({}, 0.0)

    if not exact_match:
        mcs_graph = rdFMCS.FindMCS(
            [template, query],
            bondCompare=rdFMCS.BondCompare.CompareAny,
            atomCompare=rdFMCS.AtomCompare.CompareAny,
            timeout=40,
            completeRingsOnly=True,
        )
    else:
        mcs_graph = rdFMCS.FindMCS(
            [template, query],
            bondCompare=rdFMCS.BondCompare.CompareOrderExact,
            atomCompare=rdFMCS.AtomCompare.CompareElements,
            timeout=40,
            completeRingsOnly=True,
        )

    substructure, sim_score = _generate_sim_score(
        template, query, mcs_graph.smartsString
    )

    return ParityResult(substructure, sim_score)



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Generic helper functions that may be re-used
"""


def find_element_in_list(array, element):
    """
    Finds an element in an array. Does not crash if not found

    Args:
        array (list): list to be searched
        element (any): element to be found

    Returns:
        int: Index of the element or None if the element is not found.
    """
    try:
        index = array.index(element)
        return index
    except ValueError:
        return None



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2019 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Module housing some of the dataclasses used throughout the
pdbeccdutils application.
"""

from dataclasses import dataclass
from datetime import date
from enum import IntEnum
from typing import Any, List, NamedTuple, Dict
import rdkit


class DepictionSource(IntEnum):
    """Where does the depiction come from.

    Attributes:
        Pubchem - Pubchem layout used
        Template - general substructure used
        RDKit - RDKit functionality using Coordgen.
        Failed - Nothing worked.
    """

    PubChem = 1
    Template = 2
    RDKit = 3
    Failed = 4


class ConformerType(IntEnum):
    """Conformer type of the `Component` object.

    Attributes:
        Ideal
        Model
        Depiction: 2D conformation
        Computed
        AllConformers
    """

    Ideal = 1
    Model = 2
    Depiction = 3
    Computed = 4
    AllConformers = 5


class ReleaseStatus(IntEnum):
    """An enumeration for pdbx_release_status
    allowed values include REL and HOLD, see:
    http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx.dic/Items/_chem_comp.pdbx_release_status.html

    Notes:
        An additional value 'NOT_SET' has been added for case where
        pdbx_release_status has not been set.
    """

    NOT_SET = 0
    DEL = 1
    HOLD = 2
    HPUB = 3
    OBS = 4
    REF_ONLY = 5
    REL = 6

    @staticmethod
    def from_str(s):
        """Convert wwPDB CIF CCD representation to enum

        Args:
            s (str): str representation of the release status

        Returns:
            ReleaseStatus: Component release status
        """

        if s.upper() == "DEL":
            return ReleaseStatus.DEL
        if s.upper() == "HOLD":
            return ReleaseStatus.HOLD
        if s.upper() == "HPUB":
            return ReleaseStatus.HPUB
        if s.upper() == "OBS":
            return ReleaseStatus.OBS
        if s.upper() == "REF_ONLY":
            return ReleaseStatus.REF_ONLY
        if s.upper() == "REL":
            return ReleaseStatus.REL

        return ReleaseStatus.NOT_SET


class ScaffoldingMethod(IntEnum):
    """
    Rdkit scaffold methods
    """

    MurckoScaffold = 1
    MurckoGeneric = 2
    Brics = 3


class DepictionResult(NamedTuple):
    """
    Depictions result details.

    Args:
        source (DepictionSource): Source of the depiction.
        template_name (str): template name.
        mol (rdkit.Chem.rdchem.Mol): RDKit mol object.
        score (float): Quality of the depiction, lower is better.

    """

    source: DepictionSource
    template_name: str
    mol: rdkit.Chem.rdchem.Mol
    score: float


class Descriptor(NamedTuple):
    """
    Descriptor obtained from the cif file. This is essentially
    _pdbx_chem_comp_descriptor field.

    Args:
        type (str): `_pdbx_chem_comp_descriptor.type` in CIF language.
        program (str): `_pdbx_chem_comp_descriptor.program` in CIF language.
        value (str): `_pdbx_chem_comp_descriptor.descriptor` in CIF language.
    """

    type: str
    program: str
    value: str


@dataclass
class ParityResult:
    """
    NamedTuple for the result of parity method along with the details
    necessary for calculating the similarity score.

    Attributes:
        mapping (dict of str: str): Atom-level mapping template->query.
        similarity_score (float): Calculate similarity score.
    """

    mapping: Dict[str, str]
    similarity_score: float


@dataclass
class CCDProperties:
    """
    Properties of the component comming from _chem_comp namespace.

    Args:
        id (str): _chem_comp.id
        name (str): _chem_comp.name
        formula (str): _chem_comp.formula
        modified_date (date): _chem_comp.pdbx_modified_date
        pdbx_release_status (str): _chem_comp.pdbx_release_status
        weight (str): _chem_comp.formula_weight
    """

    id: str
    name: str
    formula: str
    modified_date: date
    pdbx_release_status: ReleaseStatus
    weight: float


@dataclass
class FragmentEntry:
    """Fragment entry in the fragment library

    Args:
        name (str): Name or id of the fragment.
        source (str): where does this fragment come from.
        mol (rdkit.Chem.rdchem.Mol): rdkit mol object with the fragment.
    """

    name: str
    source: str
    mol: rdkit.Chem.rdchem.Mol


@dataclass
class SubstructureMapping:
    """Represents a fragment hit in the component

    Args:
        name (str): Name of the substructure.
        smiles (str): SMILES representation of the substructure
        source (str): Where does this fragment come from.
        mapping (List[List[Any]]): Mappings with atom names or indices.
    """

    name: str
    smiles: str
    source: str
    mappings: List[List[Any]]


class BoundMolecule:
    def __init__(self, graph):
        self.graph = graph

    def to_dict(self):
        """Return dictionary style representation of the bound molecule.

        Returns:
            dict of `str`: dict representation of the object
        """

        nodes = sorted(self.graph, key=lambda l: (int(l.res_id), l.chain))

        results_bag = {"residues": [], "connections": []}
        results_bag["residues"] = [x.to_dict() for x in nodes]
        results_bag["connections"] = [
            [e.id, f.id, a] for e, f, a in self.graph.edges(data=True)
        ]

        return results_bag

    def to_list(self):
        results_bag = [
            (e.to_dict(), f.to_dict(), a) for e, f, a in self.graph.edges(data=True)
        ]
        return results_bag

    def to_arpeggio(self):
        """Return str representation of the bound molecule.

        Returns:
            list of `str`: A list of residues found in the bound molecule.
        """
        return [x.to_arpeggio() for x in self.graph.nodes]

    def __str__(self):
        return "-".join(self.to_arpeggio())


class Residue:
    """Represents a single residue."""

    def __init__(self, name, chain, res_id, ins_code, ent_id):
        self.name = name
        self.chain = chain
        self.res_id = res_id
        self.ins_code = "" if not ins_code else ins_code
        self.ent_id = ent_id
        self.id = f"{chain}{res_id}{self.ins_code}"

    def __eq__(self, other):
        if self.id != other.id:
            return False

        return True

    def to_dict(self):
        """Returns a dictionary representation of a given residue.

        Returns:
            (:obj:`dict` of :obj:`str`): Dictionary representation along
            with the mmCIF keys.
        """
        return {
            "id": self.id,
            "label_comp_id": self.name,
            "auth_asym_id": self.chain,
            "auth_seq_id": self.res_id,
            "pdbx_PDB_ins_code": " " if not self.ins_code else self.ins_code,
            "entity_id": self.ent_id,
        }

    def __hash__(self):
        return hash(f"{self.chain}{self.res_id}{self.name}{self.ins_code}")

    def __str__(self):
        return f"/{self.name}/{self.res_id}{self.ins_code}/{self.chain}/"

    def to_arpeggio(self):
        """Gets Arpeggio style representation of a residue e.g. `/A/129/`
        or /A/129A/ in case there is an insertion code.

        Returns:
            str: Residue description in Arpeggio style.
        """
        return f"/{self.chain}/{self.res_id}{self.ins_code}/"



import pytest
from rdkit import Chem

from pdbeccdutils.core import ccd_reader
from pdbeccdutils.core.models import ScaffoldingMethod
from pdbeccdutils.tests.tst_utilities import cif_filename

test_inputs = [
    ("IBP", "c1ccccc1"),
    ("NAG", "C1CCOCC1"),
    ("VIA", "O=c1[nH]c(-c2cccc(S(=O)(=O)N3CCNCC3)c2)nc2cn[nH]c12"),
    ("007", "c1ccc(C2CCCC2)cc1"),
    (
        "BCD",
        "C1C[C@@H]2OC[C@H]1O[C@@H]1CC[C@@H](CO1)O[C@@H]1CC[C@@H](CO1)O[C@@H]1CC[C@@H](CO1)O[C@@H]1CC[C@@H](CO1)O[C@@H]1CC[C@@H](CO1)O[C@@H]1CC[C@@H](CO1)O2",
    ),
]


class TestScaffold:
    """
    Test of the Scaffolding generation
    """

    @staticmethod
    @pytest.mark.parametrize("ccd_id,smiles", test_inputs)
    def test_scaffold_present(ccd_id, smiles):
        component = ccd_reader.read_pdb_cif_file(cif_filename(ccd_id)).component
        result = component.get_scaffolds()

        mol = result[0]

        assert Chem.MolToSmiles(mol) == smiles
        assert component.scaffolds
        assert component.scaffolds[0].smiles == smiles

    @staticmethod
    @pytest.mark.parametrize("ccd_id", ["EOH", "DMS"])
    def test_scaffold_absent(ccd_id):
        component = ccd_reader.read_pdb_cif_file(cif_filename(ccd_id)).component
        result = component.get_scaffolds()

        assert result[0].GetNumAtoms() == 0
        assert not component.scaffolds

    @staticmethod
    @pytest.mark.parametrize(
        "scaffold_type", [ScaffoldingMethod.MurckoGeneric, ScaffoldingMethod.Brics]
    )
    def test_scaffolds(scaffold_type):
        c = ccd_reader.read_pdb_cif_file(cif_filename("NAG")).component
        c.get_scaffolds(scaffold_type)

        assert len(c.scaffolds) > 0



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.


import os

import numpy
import requests
from pdbeccdutils.core import ccd_reader
from rdkit import Chem
from rdkit.Chem import AllChem

import logging

logging.getLogger("urllib3").setLevel(logging.WARNING)


def rescale_molecule(path, factor):
    """
    Rescale molecule coords to a given factor

    Args:
        path (str) Path to the molecule to be rescaled.
        factor (float): rescaling factor
    """
    mol = Chem.MolFromMolFile(path, sanitize=True)
    matrix = numpy.zeros((4, 4), float)

    for i in range(3):
        matrix[i, i] = factor
    matrix[3, 3] = 1

    AllChem.TransformMol(mol, matrix)
    Chem.MolToMolFile(mol, path)


def download_template(destination: str, template_id: str, inchikey: str) -> bool:
    """Download 2D layout from the PubChem FTP.

    Args:
        destination (str): Path to the pubchem template
        template_id (str): CCD id of a pubchem template
        inchikey (str): CCD's INCHIKey

    Returns:
        bool: If the download was successful or no.
    """
    pubchem_api = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound"

    if os.path.isfile(destination):
        return False

    inchi_url = f"{pubchem_api}/inchikey/{inchikey}/cids/json"
    response = requests.get(inchi_url)

    if response.status_code != 200:
        return False

    json_file = response.json()
    cid = json_file["IdentifierList"]["CID"][0]

    structure_url = f"{pubchem_api}/cid/{cid}/record/SDF/?record_type=2d&response_type=save&response_basename={template_id}.sdf"
    response = requests.get(structure_url)

    if response.status_code != 200:
        return False

    with open(destination, "wb") as fp:
        fp.write(response.content)

    return True


class PubChemDownloader:
    """
    Toolkit to retrieve 2D layouts from the PubChem database.
    """

    def __init__(self, pubchem_templates: str) -> None:
        if not os.path.isdir(pubchem_templates):
            raise ValueError(f"{pubchem_templates} is not a valid path")

        self.pubchem_templates = pubchem_templates

    def update_ccd_dir(self, components: str):
        """Update 2D images of PDBeChem components which are available
        in the pubchem database

        Args:
            components (str): Path to the directory with components in
                .cif format
        """

        for f in os.listdir(components):
            c = ccd_reader.read_pdb_cif_file(os.path.join(components, f)).component
            self.process_template(c)

    def update_ccd_file(self, ccd: str) -> None:
        """Update 2d images of pdbechem components which are available
        in the pubchem database from CCD files.

        Args:
            ccd (str): Path to the the `.cif` CCD file
        """
        components = ccd_reader.read_pdb_components_file(ccd)

        for i in components.values():
            self.process_template(i.component)

    def process_template(self, component):
        """Process template for a given component. First the component
        is attempted to be downloaded and re-scaled. Since the RDKit
        default depiction has 1.5A single bond size whereas templates
        from pubchem are 1.0A.

        Args:
            component (Component): Component
            destination (str): Path to the pubchem 2D template dir

        Returns:
            bool: whether or not the new template has been processed
        """
        destination = os.path.join(self.pubchem_templates, f"{component.id}.sdf")
        downloaded = download_template(destination, component.id, component.inchikey)

        if downloaded:
            rescale_molecule(destination, 1.5)

        return downloaded



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.


def str_to_int(i):
    """
    Converts a string into integer. Returns 0 if a string cannot be
    converted.

    Args:
        i (str): string representation of an integer

    Returns:
        int: conversion of the string
    """
    try:
        return int(i)
    except ValueError:
        return 0


def str_to_float(f):
    """
    Converts a string into float. Returns 0.0 if a string cannot be
    converted.

    Args:
        i (str): string representation of a float

    Returns:
        float: conversion of the string
    """
    try:
        return float(f)
    except ValueError:
        return 0.0


def listit(t):
    """Format deep tuples into deep list

    Args:
        t (tuple of tuples): deep tuples

    Returns:
        list[list]: deep list
    """
    return list(map(listit, t)) if isinstance(t, (list, tuple)) else t



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""
A set of methods for reading in data and creating internal representation
of molecules. The basic use can be as easy as this:

    from pdbeccdutils.core import ccd_reader

    ccdutils_component = ccd_reader.read_pdb_cif_file('/path/to/cif/ATP.cif').component
    rdkit_mol = ccdutils_component.mol
"""

import logging
import os
from datetime import date
from typing import Dict, List, NamedTuple

import rdkit
from pdbeccdutils.core.component import Component
from pdbeccdutils.core.exceptions import CCDUtilsError
from pdbeccdutils.core.models import (
    CCDProperties,
    ConformerType,
    Descriptor,
    ReleaseStatus,
)
from pdbeccdutils.helpers import cif_tools, conversions, mol_tools
from gemmi import cif

# categories that need to be 'fixed'
# str => list[str]
preprocessable_categories = [
    "_chem_comp_atom.",
    "_chem_comp_bond.",
    "_pdbx_chem_comp_identifier.",
    "_pdbx_chem_comp_descriptor.",
    "_chem_comp.",
]


class CCDReaderResult(NamedTuple):
    """
    NamedTuple for the result of reading an individual PDB chemical
    component definition (CCD).

    Attributes:
        component (Component): internal representation of the CCD read-in.
        errors (list[str]): A list of any errors
            found while reading the CCD. If no warnings found `errors`
            will be empty.
        warnings (list[str]): A list of any warnings
            found while reading the CCD. If no warnings found `warnings`
            will be empty.
        sanitized (bool): Whether or not the molecule was sanitized
    """

    warnings: List[str]
    errors: List[str]
    component: Component
    sanitized: bool


def read_pdb_cif_file(path_to_cif: str, sanitize: bool = True) -> CCDReaderResult:
    """
    Read in single wwPDB CCD CIF component and create its internal
    representation.

    Args:
        path_to_cif (str): Path to the cif file
        sanitize (bool): [Defaults: True]

    Raises:
        ValueError: if file does not exist

    Returns:
        CCDReaderResult: Results of the parsing altogether with
        the internal representation of the component.
    """
    if not os.path.isfile(path_to_cif):
        raise ValueError("File '{}' does not exists".format(path_to_cif))

    doc = cif.read(path_to_cif)
    cif_block = doc.sole_block()

    return _parse_pdb_mmcif(cif_block, sanitize)


def read_pdb_components_file(
    path_to_cif: str, sanitize: bool = True
) -> Dict[str, CCDReaderResult]:
    """
    Process multiple compounds stored in the wwPDB CCD
    `components.cif` file.

    Args:
        path_to_cif (str): Path to the `components.cif` file with
            multiple ligands in it.
        sanitize (bool): Whether or not the components should be sanitized
            Defaults to True.

    Raises:
        ValueError: if the file does not exist.

    Returns:
        dict[str, CCDReaderResult]: Internal representation of all
        the components in the `components.cif` file.
    """
    if not os.path.isfile(path_to_cif):
        raise ValueError("File '{}' does not exists".format(path_to_cif))

    result_bag = {}

    for block in cif.read(path_to_cif):
        try:
            result_bag[block.name] = _parse_pdb_mmcif(block, sanitize)
        except CCDUtilsError as e:
            logging.error(
                f"ERROR: Data block {block.name} not processed. Reason: ({str(e)})."
            )

    return result_bag


# region parse mmcif


def _parse_pdb_mmcif(cif_block, sanitize=True):
    """
    Create internal representation of the molecule from mmcif format.

    Args:
        cif_block (cif.Block): mmcif block object from gemmi
        sanitize (bool): Whether or not the rdkit component should
            be sanitized. Defaults to True.

    Returns:
        CCDReaderResult: internal representation with the results
            of parsing and Mol object.
    """
    warnings = []
    errors = []
    sanitized = False
    mol = rdkit.Chem.RWMol()

    for c in preprocessable_categories:
        w = cif_tools.preprocess_cif_category(cif_block, c)

        if w:
            warnings.append(w)

    _parse_pdb_atoms(mol, cif_block)
    _parse_pdb_conformers(mol, cif_block)
    _parse_pdb_bonds(mol, cif_block, errors)
    _handle_implicit_hydrogens(mol)

    if sanitize:
        sanitized = mol_tools.sanitize(mol)

    descriptors = _parse_pdb_descriptors(
        cif_block, "_pdbx_chem_comp_descriptor.", "descriptor"
    )
    descriptors += _parse_pdb_descriptors(
        cif_block, "_pdbx_chem_comp_identifier.", "identifier"
    )
    properties = _parse_pdb_properties(cif_block)

    comp = Component(mol.GetMol(), cif_block, properties, descriptors)
    reader_result = CCDReaderResult(
        warnings=warnings, errors=errors, component=comp, sanitized=sanitized
    )

    return reader_result


def _parse_pdb_atoms(mol, cif_block):
    """
    Setup atoms in the component

    Args:
        mol (rdkit.Chem.rchem.Mol): Rdkit Mol object with the
            compound representation.
        cif_block (cif.Block): mmCIF block object from gemmi.
    """
    if "_chem_comp_atom." not in cif_block.get_mmcif_category_names():
        return

    atoms = cif_block.find(
        "_chem_comp_atom.",
        ["atom_id", "type_symbol", "alt_atom_id", "pdbx_leaving_atom_flag", "charge"],
    )
    for row in atoms:
        atom_id = row["_chem_comp_atom.atom_id"].strip('"')
        element = row["_chem_comp_atom.type_symbol"].strip('"')
        alt_atom_id = row["_chem_comp_atom.alt_atom_id"].strip('"')
        leaving_atom = row["_chem_comp_atom.pdbx_leaving_atom_flag"].strip('"')
        charge = row["_chem_comp_atom.charge"].strip('"')

        element = element if len(element) == 1 else element[0] + element[1].lower()
        isotope = None

        if element == "D":
            element = "H"
            isotope = 2
        elif element == "X":
            element = "*"

        atom = rdkit.Chem.Atom(element)
        atom.SetProp("name", atom_id)
        atom.SetProp("alt_name", alt_atom_id)
        atom.SetBoolProp("leaving_atom", leaving_atom == "Y")
        atom.SetFormalCharge(conversions.str_to_int(charge))

        if isotope is not None:
            atom.SetIsotope(isotope)

        mol.AddAtom(atom)


def _parse_pdb_conformers(mol, cif_block):
    """Setup model and ideal cooordinates in the rdkit Mol object.

    Args:
        mol (rdkit.Chem.rdchem.Mol): RDKit Mol object with the compound
            representation.
        cif_block (cif.Block): mmCIF block object from gemmi.
    """

    if "_chem_comp_atom." not in cif_block.get_mmcif_category_names():
        return

    required_fields = [
        "model_Cartn_x",
        "model_Cartn_y",
        "model_Cartn_z",
        "pdbx_model_Cartn_x_ideal",
        "pdbx_model_Cartn_y_ideal",
        "pdbx_model_Cartn_z_ideal",
    ]
    atoms = cif_block.find("_chem_comp_atom.", required_fields)
    ideal = _setup_pdb_conformer(
        atoms, "pdbx_model_Cartn_{}_ideal", ConformerType.Ideal.name
    )
    model = _setup_pdb_conformer(atoms, "model_Cartn_{}", ConformerType.Model.name)

    mol.AddConformer(ideal, assignId=True)
    mol.AddConformer(model, assignId=True)


def _setup_pdb_conformer(atoms, label, name):
    """Setup a conformer

    Args:
        atoms (Table): mmcif table with the atom info.
        label (str): Conformer category
        name (str): Conformer name.

    Returns:
        rdkit.Chem.rdchem.Conformer: Conformer of the component.
    """
    if not atoms:
        return

    conformer = rdkit.Chem.Conformer(len(atoms))

    for row in atoms:
        x = conversions.str_to_float(
            row["_chem_comp_atom.{}".format(label.format("x"))]
        )
        y = conversions.str_to_float(
            row["_chem_comp_atom.{}".format(label.format("y"))]
        )
        z = conversions.str_to_float(
            row["_chem_comp_atom.{}".format(label.format("z"))]
        )

        atom_position = rdkit.Chem.rdGeometry.Point3D(x, y, z)
        conformer.SetAtomPosition(row.row_index, atom_position)

    conformer.SetProp("name", name)

    return conformer


def _parse_pdb_bonds(mol, cif_block, errors):
    """
    Setup bonds in the compound

    Args:
        mol (rdkit.Chem.rdchem.Mol): Molecule which receives bonds.
        cif_block (cif.Block): mmcif block Block from gemmi.
        errors (list[str]): Issues encountered while parsing.
    """
    if (
        "_chem_comp_atom." not in cif_block.get_mmcif_category_names()
        or "_chem_comp_bond." not in cif_block.get_mmcif_category_names()
    ):
        return
    atoms = cif_block.find("_chem_comp_atom.", ["atom_id"])
    bonds = cif_block.find(
        "_chem_comp_bond.", ["atom_id_1", "atom_id_2", "value_order"]
    )
    atoms_ids = list(atoms.find_column("_chem_comp_atom.atom_id"))
    for row in bonds:
        try:
            atom_1 = row["_chem_comp_bond.atom_id_1"]
            atom_1_id = atoms_ids.index(atom_1)
            atom_2 = row["_chem_comp_bond.atom_id_2"]
            atom_2_id = atoms_ids.index(atom_2)
            bond_order = _bond_pdb_order(row["_chem_comp_bond.value_order"])

            mol.AddBond(atom_1_id, atom_2_id, bond_order)
        except ValueError:
            errors.append(
                f"Error perceiving {atom_1} - {atom_2} bond in _chem_comp_bond"
            )
        except RuntimeError:
            errors.append(f"Duplicit bond {atom_1} - {atom_2}")


def _handle_implicit_hydrogens(mol):
    """Forbid atoms which does not have explicit Hydrogen partner to get
    implicit hydrogen.

    Args:
        mol (rkit.Chem.rdchem.Mol): Mol to be modified
    """
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() == 1:
            continue

        no_Hs = True
        for bond in atom.GetBonds():
            other = bond.GetOtherAtom(atom)
            if other.GetAtomicNum() == 1:
                no_Hs = False
                break

        atom.SetNoImplicit(no_Hs)


def _parse_pdb_descriptors(cif_block, cat_name, label="descriptor"):
    """Parse useful information from _pdbx_chem_comp_* category

    Args:
        cif_block (cif.Block): mmCIF Block object from gemmi
        cat_name (str): mmcif category with the
            descriptors info.
        label (str, optional): Defaults to 'descriptor'. Name of the
            category to be parsed.

    Returns:
        Descriptor: namedtuple with the property info
    """
    descriptors = []

    if cat_name not in cif_block.get_mmcif_category_names():
        return descriptors

    descriptors_block = cif_block.find(cat_name, [label, "type", "program"])
    for row in descriptors_block:
        d = Descriptor(
            type=row[f"{cat_name}type"],
            program=row[f"{cat_name}program"],
            value=row[f"{cat_name}{label}"],
        )
        descriptors.append(d)

    return descriptors


def _parse_pdb_properties(cif_block):
    """Parse useful information from _chem_comp category

    Args:
        cif_block (cif.Block): mmcif block object from gemmi

    Returns:
        Properties: dataclass with the CCD properties.
    """
    properties = None
    if "_chem_comp." in cif_block.get_mmcif_category_names():
        mod_date = cif_block.find_value("_chem_comp.pdbx_modified_date")
        if cif.is_null(mod_date):
            d = date(1970, 1, 1)
        else:
            mod_date = mod_date.split("-")
            d = date(int(mod_date[0]), int(mod_date[1]), int(mod_date[2]))

        rel_status = ReleaseStatus.from_str(
            cif_block.find_value("_chem_comp.pdbx_release_status")
        )
        formula_weight = cif_block.find_value("_chem_comp.formula_weight")
        weight = 0.0 if cif.is_null(formula_weight) else cif.as_number(formula_weight)

        properties = CCDProperties(
            id=cif_block.find_value("_chem_comp.id"),
            name=cif_block.find_value("_chem_comp.name").strip('"'),
            formula=cif_block.find_value("_chem_comp.formula").strip('"'),
            modified_date=d,
            pdbx_release_status=rel_status,
            weight=weight,
        )
    return properties


def _bond_pdb_order(value_order):
    """
    Transpils mmcif bond order into rdkit language

    Args:
        value_order (str): bond type as a str

    Returns:
        rdkit.Chem.rdchem.BondType: -- bond type
    """
    if value_order == "SING":
        return rdkit.Chem.rdchem.BondType(1)
    if value_order == "DOUB":
        return rdkit.Chem.rdchem.BondType(2)
    if value_order == "TRIP":
        return rdkit.Chem.rdchem.BondType(3)

    return None


# endregion parse mmcif



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.


"""Module to aid generation of 2D depictions and evaluation of their
quality

"""
import logging
import math
import os
import sys
from collections import OrderedDict
from typing import Dict

import rdkit
from pdbeccdutils.core.models import DepictionResult, DepictionSource
from pdbeccdutils.helpers.mol_tools import fix_conformer
from pdbeccdutils.utils import config
from rdkit import Chem, Geometry
from rdkit.Chem import AllChem, rdCoordGen
from scipy.spatial import KDTree


class DepictionManager:
    """
    Toolkit for depicting ligand's structure using RDKit.
    One can supply either templates or 2D depictions by pubchem.
    PubChem templates can be downloaded using PubChemDownloader class.
    """

    def __init__(
        self,
        pubchem_templates_path: str = "",
        general_templates_path: str = config.general_templates,
    ) -> None:
        """
        Initialize component which does the ligand depiction.
        If Nones is provided as parameters just the defalt RDKit
        functionality is going to be used.

        Args:
            pubchem_templates_path (str, optional): Defaults to ''.
                Path to the library with 2D structures downloaded from
                PubChem. Use `setup_pubchem_library` for this task.
            general_templates_path (str, optional): Defaults to
                config.general_templates (supplied with the pdbeccdutils).
                Path to the library with general templates to be used
                for depicting ligand e.g. porphyrin rings.

        """
        self.coordgen_params = rdCoordGen.CoordGenParams()
        self.coordgen_params.coordgenScaling = 50 / 1.5
        self.coordgen_params.templateFileDir = config.coordgen_templates

        self.pubchem_templates = (
            pubchem_templates_path if os.path.isdir(pubchem_templates_path) else ""
        )
        self.templates: Dict[str, rdkit.Chem.rdchem.Mol] = OrderedDict()

        if os.path.isdir(general_templates_path):
            for k in sorted(os.listdir(general_templates_path)):
                template = self._load_template(os.path.join(general_templates_path, k))
                template_name = k.split(".")[0]
                self.templates[template_name] = template

    def depict_molecule(
        self, het_id: str, mol: rdkit.Chem.rdchem.Mol
    ) -> DepictionResult:
        """
        Given input molecule tries to generate its depictions.

        Presently 3 methods are used:
                Pubchem template - find 2d depiction in pubchem db
                User-provided templates - try to use general templates
                From 3D conformer - just apply default RDKit functionality

        Arguments:
            id (str): id of the ligand
            mol (rdkit.Chem.rdchem.Mol): molecule to be depicted

        Returns:
            DepictionResult: Summary of the ligand depiction process.
        """
        temp_mol = Chem.RWMol(mol)

        templateMol = Chem.RWMol(temp_mol).GetMol()
        pubchemMol = Chem.RWMol(temp_mol).GetMol()
        rdkitMol = Chem.RWMol(temp_mol).GetMol()
        results = []

        pubchem_res = (
            self._get_2D_by_pubchem(het_id, pubchemMol)
            if self.pubchem_templates
            else None
        )
        template_res = self._get_2D_by_template(templateMol) if self.templates else []
        rdkit_res = self._get_2D_by_rdkit(rdkitMol)

        if pubchem_res is not None:
            results.append(pubchem_res)
        if rdkit_res is not None:
            results.append(rdkit_res)

        results = results + template_res

        results.sort(key=lambda l: (l.score, l.source))

        if results:
            to_return = results[0]
            fix_conformer(to_return.mol.GetConformer(0))

            return to_return

        return DepictionResult(
            source=DepictionSource.Failed, template_name="", mol=None, score=1000
        )

    def _get_pubchem_template_path(self, het_id):
        """Get path to the PubChem template if it exists.

        Args:
            het_id (str): Ligand in.

        Returns:
            str: Path to the PubChem layout.
        """
        path = os.path.join(self.pubchem_templates, f"{het_id}.sdf")

        return path if os.path.isfile(path) else ""

    def _load_template(self, path):
        """
        Loads a template molecule with 2D coordinates

        Args:
            path (str): path to the model molecule in *.sdf,
                or *.pdb format

        Raises:
            ValueError: if unsupported format is used: sdf|pdb

        Returns:
            rdkit.Chem.rdchem.Mol: RDKit representation of the template
        """
        mol = Chem.RWMol()
        extension = os.path.basename(path).split(".")[1]

        if extension == "sdf":
            mol = Chem.MolFromMolFile(path, sanitize=True, removeHs=True)
        elif extension == "pdb":
            mol = Chem.MolFromPDBFile(path, sanitize=True, removeHs=True)
        else:
            raise ValueError("Unsupported molecule type '{}'".format(extension))

        p = Chem.AdjustQueryParameters()
        p.makeAtomsGeneric = True
        p.makeBondsGeneric = True

        mol = Chem.AdjustQueryProperties(mol, p)

        return mol

    def _get_2D_by_rdkit(self, mol):
        """
        Get depiction done using solely the default RDKit functionality.

        Args:
            mol (rdkit.Chem.rdchem.Mol): Mol to be depicted

        Returns:
            DepictionResult: Depiction with some usefull metadata
        """
        try:
            rdCoordGen.AddCoords(mol, self.coordgen_params)
            flaws = DepictionValidator(mol).depiction_score()
            return DepictionResult(
                source=DepictionSource.RDKit, template_name=None, mol=mol, score=flaws
            )
        except Exception:
            return DepictionResult(
                source=DepictionSource.Failed, template_name=None, mol=None, score=1000
            )

    def _get_2D_by_pubchem(self, code, mol):
        """
        Depict ligand using Pubchem templates.

        Args:
            code (str): id of the molecule to be processed
            mol (rdkit.Chem.rdchem.Mol): Mol to be depicted

        Returns:
            DepictionResult: Depiction with some usefull metadata
        """
        try:
            template_path = self._get_pubchem_template_path(code)
            if template_path:
                template = self._load_template(template_path)

                if mol.HasSubstructMatch(template):
                    AllChem.GenerateDepictionMatching2DStructure(mol, template)
                    flaws = DepictionValidator(mol).depiction_score()
                    return DepictionResult(
                        source=DepictionSource.PubChem,
                        template_name=code,
                        mol=mol,
                        score=flaws,
                    )
        except Exception as e:
            print(str(e), file=sys.stderr)

        return DepictionResult(
            source=DepictionSource.Failed, template_name=None, mol=None, score=1000
        )

    def _get_2D_by_template(self, mol):
        """
        Depict ligand using user-provided templates

        Args:
            mol (rdkit.Chem.rchem.Mol): Mol to be depicted

        Returns:
            :obj:`list` of :obj:`DepictionResult`: Depictions with their
            quality and metadata.
        """
        results = list()
        try:
            for key, template in self.templates.items():
                temp_mol = Chem.RWMol(mol)
                if temp_mol.HasSubstructMatch(template):
                    AllChem.GenerateDepictionMatching2DStructure(temp_mol, template)
                    flaws = DepictionValidator(temp_mol).depiction_score()
                    results.append(
                        DepictionResult(
                            source=DepictionSource.Template,
                            template_name=key,
                            mol=temp_mol,
                            score=flaws,
                        )
                    )
        except Exception:
            # if it fails it fails, but generally it wont
            logging.warning("Depiction generation by template failed")

        return results


class DepictionValidator:
    """
    Toolkit for estimation of depiction quality
    """

    def __init__(self, mol):
        self.mol = mol
        self.conformer = mol.GetConformer()
        self.bonds = self.mol.GetBonds()

        atoms = [
            self.conformer.GetAtomPosition(i)
            for i in range(0, self.conformer.GetNumAtoms())
        ]
        atom_centers = [[atom.x, atom.y, atom.z] for atom in atoms]

        self.kd_tree = KDTree(atom_centers)

    def _intersection(self, bondA, bondB):
        """
        True if two bonds collide, false otherwise. Note that false is
        retrieved even in case the bonds share common atom, as this is
        not a problem case. Cramer's rule is used for the linear
        equations system.

        Args:
            bondA (rdkit.Chem.rdchem.Bond): this bond
            bondB (rdkit.Chem.rdchem.Bond): other bond

        Returns:
            bool: true if bonds share collide, false otherwise.
        """
        atoms = [
            bondA.GetBeginAtom(),
            bondA.GetEndAtom(),
            bondB.GetBeginAtom(),
            bondB.GetEndAtom(),
        ]
        names = [a.GetProp("name") for a in atoms]
        points = [self.conformer.GetAtomPosition(a.GetIdx()) for a in atoms]

        vecA = Geometry.Point2D(points[1].x - points[0].x, points[1].y - points[0].y)
        vecB = Geometry.Point2D(points[3].x - points[2].x, points[3].y - points[2].y)

        # we need to set up directions of the vectors properly in case
        # there is a common atom. So we identify angles correctly
        # e.g. B -> A; B -> C and not A -> B; C -> B.
        if len(set(names)) == 3:
            angle = self.__get_angle(names, vecA, vecB)
            return angle < 10.0

        # Cramer's rule to identify intersection
        det = vecA.x * -vecB.y + vecA.y * vecB.x
        if round(det, 2) == 0.00:
            return False

        a = points[2].x - points[0].x
        b = points[2].y - points[0].y

        detP = (a * -vecB.y) - (b * -vecB.x)
        p = round(detP / det, 3)

        if p < 0 or p > 1:
            return False

        detR = (vecA.x * b) - (vecA.y * a)
        r = round(detR / det, 3)

        if 0 <= r <= 1:
            return True

        return False

    def __get_angle(self, names, vecA, vecB):
        """Get the size of the angle formed by two bonds which share
        common atom.

        Args:
            names (list of str): List of atom names forming bonds
                [A, B, C, D] for AB and CD.
            vecA (Geometry.Point2D): Vector representing AB bond.
            vecB (Geometry.Point2D): Vector representing CD bond.

        Returns:
            float: Size of the angle in degrees.
        """
        pivot = max(names, key=names.count)

        if names[0] != pivot:  # Atoms needs to be order to pick vectors correctly
            vecA = vecA * -1

        if names[2] != pivot:
            vecB = vecB * -1

        radians = vecA.AngleTo(vecB)
        angle = 180 / math.pi * radians

        return angle

    def has_degenerated_atom_positions(self, threshold):
        """
        Detects whether the structure has a pair or atoms closer to each
        other than threshold. This can detect structures which may need
        a template as they can be handled by RDKit correctly.

        Arguments:
            threshold (float): Bottom line to use for spatial search.

        Returns:
            (bool): if such atomic pair is found
        """

        for i in range(0, len(self.conformer.GetNumAtoms())):
            center = self.conformer.GetAtomPosition(i)
            point = [center.x, center.y, center.z]
            surrounding = self.kd_tree.query_ball_point(point, threshold)

            if len(surrounding) > 1:
                return True

        return False

    def count_suboptimal_atom_positions(self, lower_bound, upper_bound):
        """
        Detects whether the structure has a pair or atoms in the range
        <lowerBound, upperBound> meaning that the depiction could
        be improved.

        Arguments:
            lower_bound (float): lower bound
            upper_bound (float): upper bound

        Returns:
            float: number of atoms with crowded neighbourhood
        """
        counter = 0
        for i in range(self.conformer.GetNumAtoms()):
            center = self.conformer.GetAtomPosition(i)
            point = [center.x, center.y, center.z]
            surrounding_low = self.kd_tree.query_ball_point(point, lower_bound)
            surrounding_high = self.kd_tree.query_ball_point(point, upper_bound)

            if len(surrounding_high) - len(surrounding_low) > 0:
                counter += 1

        return counter / 2

    def count_bond_collisions(self):
        """
        Counts number of collisions among all bonds. Can be used for estimations of how 'wrong'
        the depiction is.

        Returns:
            int: number of bond collisions per molecule
        """

        errors = 0

        for i in range(0, len(self.bonds)):
            for a in range(i + 1, len(self.bonds)):
                result = self._intersection(self.bonds[i], self.bonds[a])

                if result:
                    errors += 1
        return errors

    def has_bond_crossing(self):
        """
        Tells if the structure contains collisions

        Returns:
            bool: Indication about bond collisions
        """
        return self.count_bond_collisions() > 0

    def depiction_score(self):
        """
        Calculate quality of the ligand depiction. The higher the worse.
        Ideally that should be 0.

        Returns:
            float: Penalty score.
        """

        collision_penalty = 1
        degenerated_penalty = 0.4

        bond_collisions = self.count_bond_collisions()
        degenerated_atoms = self.count_suboptimal_atom_positions(0.0, 0.5)

        score = (
            collision_penalty * bond_collisions
            + degenerated_penalty * degenerated_atoms
        )

        return round(score, 1)



"""Unit tests for checking whether or not depictions are written properly.
"""


import json
import os
import re
import xml.etree.ElementTree as ET

import pytest
from pdbeccdutils.core import ccd_reader
from pdbeccdutils.core.component import Component
from pdbeccdutils.core.depictions import DepictionManager, DepictionSource
from pdbeccdutils.helpers.drawing import save_no_image, svg_namespace
from pdbeccdutils.tests.tst_utilities import cif_filename

collision_free_templates = [
    ("hem", ["HEM", "HEA", "HEB", "HEC", "HDD", "HEG"]),
    ("porphycene", ["HNN", "HME"]),
    ("ru_complex", ["11R"]),
    ("ru_complex", ["11R"]),
    ("phorbine", ["G2O"]),
]

collision_templates = [("cube", ["SF4", "0KA", "1CL"]), ("adamantane", ["ADM"])]

depictions = DepictionManager()


def load_molecule(id_):
    c = ccd_reader.read_pdb_cif_file(cif_filename(id_)).component
    c.compute_2d(depictions)

    return c


class TestWriteImg:
    @staticmethod
    def test_file_generated(tmpdir):  # tmpdir is a fixture with temporary directory
        mol = load_molecule("ATP")
        path = str(tmpdir.join("atp_test.svg"))
        mol.export_2d_svg(path)

        assert os.path.isfile(path)

    @staticmethod
    @pytest.mark.parametrize("ccd_id", ["NAG", "ATP", "08T", "BCD", "10R", "0OD"])
    def test_image_generation_with_names(tmpdir, ccd_id):
        mol = load_molecule(ccd_id)
        path = str(tmpdir.join(f"{ccd_id}.svg"))
        path_with_names = str(tmpdir.join(f"{ccd_id}_names.svg"))
        mol.export_2d_svg(path, names=False)
        mol.export_2d_svg(path_with_names, names=True)

        img = ET.parse(path).findall("svg:path", svg_namespace)
        img_names = ET.parse(path_with_names).findall("svg:path", svg_namespace)

        img = [
            x
            for x in img
            if "class" in x.attrib and re.fullmatch(r"atom-\d+", x.attrib["class"])
        ]
        img_names = [
            x
            for x in img_names
            if "class" in x.attrib and re.fullmatch(r"atom-\d+", x.attrib["class"])
        ]

        assert len(img_names) > len(img)

    @staticmethod
    @pytest.mark.parametrize("expected_template,names", collision_free_templates)
    def test_collision_free_template_picked(expected_template, names):
        """Test if expected templates are picked for certain compounds

        Args:
            expected_template (str): template name.
            names (list of str): list of ids matching this template.
        """
        d = DepictionManager()

        for ccd_id in names:
            mol = load_molecule(ccd_id)
            response = mol.compute_2d(d)

            assert response.source == DepictionSource.Template
            assert response.score == 0
            assert response.template_name == expected_template

    @staticmethod
    @pytest.mark.parametrize("expected_template,names", collision_templates)
    def test_collision_template_picked(expected_template, names):
        """Test if expected templates are picked for certain compounds

        Args:
            expected_template (str): template name.
            names (list of str): list of ids matching this template.
        """
        d = DepictionManager()

        for ccd_id in names:
            mol = load_molecule(ccd_id)
            response = mol.compute_2d(d)

            assert response.source == DepictionSource.Template
            assert response.score > 0
            assert response.template_name == expected_template

    @staticmethod
    def test_svg_annotation(component: Component, tmpdir_factory):
        if not component.atoms_ids:
            return

        json_obj = None
        wd = tmpdir_factory.mktemp("svg_json_test")
        out_file = os.path.join(wd, f"{component.id}.json")
        component.compute_2d(depictions)
        component.export_2d_annotation(out_file)

        assert os.path.isfile(out_file)
        assert os.path.getsize(out_file) > 0

        with open(out_file, "r") as fp:
            json_obj = json.load(fp)

        assert json_obj["ccd_id"] == component.id
        assert json_obj["resolution"]["x"] >= 0
        assert json_obj["resolution"]["y"] >= 0

        atom_names = [atom["name"] for atom in json_obj["atoms"]]
        assert len(json_obj["atoms"]) == component.mol_no_h.GetNumAtoms()
        assert len(json_obj["bonds"]) >= component.mol_no_h.GetNumBonds()
        assert all(atom["name"] for atom in json_obj["atoms"])  # do we have atom names?

        assert all(
            bond["bgn"] in atom_names and bond["end"] in atom_names
            for bond in json_obj["bonds"]
        )  # are all the atoms defined?
        assert all(
            bond["coords"] for bond in json_obj["bonds"]
        )  # do we have coordinates?
        assert all(bond["style"] for bond in json_obj["bonds"])  # and its stylling?

    @staticmethod
    def test_no_image_svg(tmpdir):
        value = "foo"
        svg = str(tmpdir.join("test.svg"))

        save_no_image(svg, default_msg="foo")

        assert os.path.isfile(svg)

        xml = ET.parse(svg)
        text = xml.find("svg:text", svg_namespace)
        assert text.text.strip() == value

    @staticmethod
    def test_no_image_png(tmpdir):
        png = str(tmpdir.join("test.png"))
        save_no_image(png)

        assert os.path.join(png)



"""Test of routine for retrieving 2D layouts of the CCDs.
"""

import os

import pytest
from rdkit import Chem

from pdbeccdutils.core import ccd_reader
from pdbeccdutils.tests.tst_utilities import cif_filename
from pdbeccdutils.utils.pubchem_downloader import PubChemDownloader

ids_to_test = ["MAN", "NAG", "SO4", "GOL", "SAC", "VIA", "GLU"]


class TestWebServices:
    @staticmethod
    @pytest.mark.parametrize("het_code", ids_to_test)
    def test_components_layouts_downloaded(tmp_path, het_code):
        dl = PubChemDownloader(str(tmp_path))

        to_download = os.path.join(tmp_path, f"{het_code}.sdf")
        comp = ccd_reader.read_pdb_cif_file(cif_filename(het_code)).component
        dl.process_template(comp)

        assert os.path.exists(to_download)
        assert os.path.getsize(to_download) > 0
        assert comp.mol_no_h.HasSubstructMatch(
            Chem.MolFromMolFile(to_download, sanitize=True)
        )

    @staticmethod
    @pytest.mark.parametrize("het_code", ids_to_test)
    def test_components_layouts_updated(tmp_path, het_code):
        dl = PubChemDownloader(str(tmp_path))

        to_download = os.path.join(tmp_path, f"{het_code}.sdf")
        dl.update_ccd_file(cif_filename(het_code))
        mol = Chem.MolFromMolFile(to_download)

        assert os.path.exists(to_download)
        assert os.path.getsize(to_download) > 0
        assert isinstance(mol, Chem.Mol)
        assert mol.GetNumAtoms() > 0

    @staticmethod
    @pytest.mark.parametrize("het_code", ids_to_test)
    def test_unichem_download(het_code):
        c = ccd_reader.read_pdb_cif_file(cif_filename(het_code)).component

        c.fetch_external_mappings()
        assert len(c.external_mappings) > 0

        c.fetch_external_mappings(True)
        assert len(c.external_mappings) > 0



import os

from setuptools import find_namespace_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise RuntimeError("Unable to find version string.")


setup(
    name="pdbeccdutils",
    version=get_version("pdbeccdutils/__init__.py"),
    description="Toolkit to deal with wwPDB chemical components definitions for small molecules.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    project_urls={
        "Source code": "https://github.com/PDBeurope/ccdutils",
        "Documentation": "https://pdbeurope.github.io/ccdutils/",
    },
    author="Protein Data Bank in Europe",
    author_email="pdbehelp@ebi.ac.uk",
    license="Apache License 2.0.",
    keywords="PDB CCD wwPDB small molecule",
    url="http://pypi.python.org/pypi/pdbeccdutils/",
    packages=find_namespace_packages(),
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=["Pillow", "scipy", "numpy", "requests", "gemmi", "networkx"],
    extras_require={
        "tests": ["pytest", "pytest-cov", "pre-commit"],
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
            "sphinx-autodoc-typehints",
            "sphinx-markdown-tables",
            "myst-parser",
        ],
    },
    entry_points={
        "console_scripts": [
            "process_components_cif=pdbeccdutils.scripts.process_components_cif_cli:main",
            "setup_pubchem_library=pdbeccdutils.scripts.setup_pubchem_library_cli:main",
        ]
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Development Status :: 5 - Production/Stable",
    ],
)



"""Common fixtures shared among all the tests
"""
from pdbeccdutils.core import ccd_writer
from pdbeccdutils.core.models import ConformerType
from pdbeccdutils.helpers.mol_tools import fix_conformer
import rdkit
from rdkit import Chem
import numpy as np
import os


def test_write_bound_molecule(boundMolecules, tmpdir_factory):
    wd = tmpdir_factory.mktemp("bm_test")
    for bm in boundMolecules:
        for remove_hs in True, False:
            suffix = f'{("" if remove_hs else "H")}'
            sdf_file = os.path.join(wd, f"{bm.id}_{suffix}.sdf")
            conf_type = ConformerType.Model
            ccd_writer.write_molecule(
                path=sdf_file,
                component=bm,
                conf_type=conf_type,
                remove_hs=remove_hs,
            )
            rdkit_mol = bm.mol_no_h if remove_hs else bm.mol
            assert os.path.isfile(sdf_file)
            assert os.path.getsize(sdf_file) > 0
            mol = Chem.MolFromMolFile(sdf_file, sanitize=False)
            assert isinstance(mol, Chem.rdchem.Mol)
            assert mol.GetNumAtoms() == rdkit_mol.GetNumAtoms()


def test_bound_molecule_has_degenerate_conformer(boundMolecules):
    for bm in boundMolecules:
        assert not bm.has_degenerated_conformer(ConformerType.Model)


def test_bound_molecule_conformer_is_broken_ion():
    mol = rdkit.Chem.RWMol()
    atom = rdkit.Chem.Atom("H")
    mol.AddAtom(atom)
    conformer = rdkit.Chem.Conformer(1)
    atom_position = rdkit.Chem.rdGeometry.Point3D(np.NaN, np.NaN, np.NaN)
    conformer.SetAtomPosition(0, atom_position)
    mol.AddConformer(conformer, assignId=True)
    m = mol.GetMol()
    c = m.GetConformer(0)
    fix_conformer(c)
    assert c.GetAtomPosition(0).x == 0.0
    assert c.GetAtomPosition(0).y == 0.0
    assert c.GetAtomPosition(0).z == 0.0


def test_bound_molecule_conformer_has_broken_atom():
    mol = rdkit.Chem.RWMol()
    o = rdkit.Chem.Atom("O")
    h = rdkit.Chem.Atom("H")
    mol.AddAtom(o)
    mol.AddAtom(h)
    mol.AddBond(0, 1, rdkit.Chem.rdchem.BondType(1))
    conformer = rdkit.Chem.Conformer(1)
    o_position = rdkit.Chem.rdGeometry.Point3D(1, 2, 3)
    h_position = rdkit.Chem.rdGeometry.Point3D(np.NaN, np.NaN, np.NaN)
    conformer.SetAtomPosition(0, o_position)
    conformer.SetAtomPosition(1, h_position)
    mol.AddConformer(conformer, assignId=True)
    m = mol.GetMol()
    c = m.GetConformer(0)
    fix_conformer(c)
    assert c.GetAtomPosition(0).x != 0.0
    assert c.GetAtomPosition(0).y != 0.0
    assert c.GetAtomPosition(0).z != 0.0
    assert c.GetAtomPosition(1).x == 0.0
    assert c.GetAtomPosition(1).y == 0.0
    assert c.GetAtomPosition(1).z == 0.0



sample_ccd_with_inchi_problems = ["7OM", "ASX", "CDL", "0OD"]


class TestIn:
    @staticmethod
    def test_inchikeys_from_rdkit_and_ccd_match(component):
        if component.id not in sample_ccd_with_inchi_problems:
            assert component.inchikey == component.inchikey_from_rdkit



import json
import os
import xml.etree.ElementTree as ET

import pytest
from pdbeccdutils.core import ccd_writer
from pdbeccdutils.core.component import Component
from pdbeccdutils.core.models import ConformerType
from gemmi import cif
from rdkit import Chem

must_have_categories = [
    "_chem_comp.",
    "_chem_comp_atom.",
    "_chem_comp_bond.",
    "_pdbx_chem_comp_descriptor.",
]


class TestFileWrites:
    @staticmethod
    def test_write_sdf(component, tmpdir_factory):
        wd = tmpdir_factory.mktemp("sdf_test")
        for ideal in True, False:
            for remove_hs in True, False:
                suffix = f'{("" if remove_hs else "H")}'
                sdf_file = os.path.join(wd, f"{component.id}_{suffix}.sdf")
                conf_type = ConformerType.Ideal if ideal else ConformerType.Model
                ccd_writer.write_molecule(
                    path=sdf_file,
                    component=component,
                    conf_type=conf_type,
                    remove_hs=remove_hs,
                )
                rdkit_mol = component.mol_no_h if remove_hs else component.mol

                assert os.path.isfile(sdf_file)
                assert os.path.getsize(sdf_file) > 0

                mol = Chem.MolFromMolFile(sdf_file, sanitize=False)

                assert isinstance(mol, Chem.rdchem.Mol)
                assert mol.GetNumAtoms() == rdkit_mol.GetNumAtoms()

    @staticmethod
    def test_pdb_write(component, tmpdir_factory):
        wd = tmpdir_factory.mktemp("pdb_test")

        if component.id in ("ASX", "NA"):
            return

        for ideal in True, False:
            for remove_h in True, False:
                suffix = "" if remove_h else "H"
                pdb_file = os.path.join(wd, f"{component.id}{suffix}.pdb")
                conformer = ConformerType.Ideal if ideal else ConformerType.Model
                ccd_writer.write_molecule(
                    pdb_file, component, remove_hs=remove_h, conf_type=conformer
                )
                rdkit_mol = component.mol_no_h if remove_h else component.mol

                assert os.path.isfile(pdb_file)
                assert os.path.getsize(pdb_file) > 0

                mol = Chem.MolFromPDBFile(
                    pdb_file,
                    removeHs=False,
                    sanitize=False,
                )

                assert isinstance(mol, Chem.rdchem.Mol)
                assert mol.GetNumAtoms() == rdkit_mol.GetNumAtoms()

    @staticmethod
    def test_pdb_fallback_write(component):
        if component.id == "ASX":
            return

        for ideal in True, False:
            for remove_h in True, False:
                conformer = ConformerType.Ideal if ideal else ConformerType.Model
                conf_id = component.get_conformer(conformer).GetId()
                rdkit_mol = component.mol_no_h if remove_h else component.mol

                pdb_repr = ccd_writer._to_pdb_str_fallback(
                    rdkit_mol, component.id, conf_id
                )

                assert pdb_repr
                mol = Chem.MolFromPDBBlock(pdb_repr, sanitize=False, removeHs=False)

                assert isinstance(mol, Chem.rdchem.Mol)
                assert mol.GetNumAtoms() == rdkit_mol.GetNumAtoms()

    @staticmethod
    def test_write_xyz(component, tmpdir_factory):
        wd = tmpdir_factory.mktemp("xyz_test")
        xyz_file = os.path.join(wd, f"{component.id}.xyz")
        ccd_writer.write_molecule(xyz_file, component)

        assert os.path.isfile(xyz_file)
        assert os.path.getsize(xyz_file) > 0

        with open(xyz_file, "r") as fp:
            lines = fp.read().splitlines()
            assert lines[1] == component.id
            assert int(lines[0]) == len(lines[2:])

    @staticmethod
    def test_xml_write(component, tmpdir_factory):
        wd = tmpdir_factory.mktemp("xml_test")
        xml_file = os.path.join(wd, f"{component.id}.xml")

        ccd_writer.write_molecule(xml_file, component)

        assert os.path.isfile(xml_file)
        assert os.path.getsize(xml_file) > 0

        xml = ET.parse(xml_file).getroot()

        assert xml.tag == "chemComp"
        assert xml.find("id").text == component.id
        assert xml.find("name").text == component.name
        assert xml.find("formula").text == component.formula
        assert xml.find("systematicName").text == next(
            (
                x.value
                for x in component.descriptors
                if x.type == "SYSTEMATIC NAME" and x.program == "ACDLabs"
            ),
            None,
        )
        assert xml.find("stereoSmiles").text == next(
            (
                x.value
                for x in component.descriptors
                if x.type == "SMILES_CANONICAL" and x.program == "CACTVS"
            ),
            None,
        )
        assert xml.find("nonStereoSmiles").text == next(
            (
                x.value
                for x in component.descriptors
                if x.type == "SMILES" and x.program == "CACTVS"
            ),
            None,
        )
        assert xml.find("inchi").text == (component.inchi if component.inchi else None)

    @staticmethod
    def test_json_write(component, tmpdir_factory):
        wd = tmpdir_factory.mktemp("json_test")

        for remove_h in True, False:
            suffix = "" if remove_h else "H"
            json_file = os.path.join(wd, f"{component.id}{suffix}.json")
            ccd_writer.write_molecule(json_file, component, remove_hs=remove_h)
            rdkit_mol = component.mol_no_h if remove_h else component.mol

            assert os.path.isfile(json_file)
            assert os.path.getsize(json_file) > 0

            with open(json_file, "r") as fp:
                js = json.load(fp)
                key = list(js.keys())[0]

                assert key == component.id
                assert len(js[key]["atoms"]) == rdkit_mol.GetNumAtoms()
                assert len(js[key]["bonds"]) == rdkit_mol.GetNumBonds()

    @staticmethod
    @pytest.mark.parametrize("rem_hs", [True, False])
    def test_cif_write(component: Component, tmpdir, rem_hs):
        path = tmpdir.join(f"{component.id}.cif")
        to_check = must_have_categories.copy()

        ccd_writer.write_molecule(str(path), component, remove_hs=rem_hs)
        cif_block = cif.read(str(path)).sole_block()

        if component.id == "NA":  # Na is an atom!
            to_check.pop(2)  # remove "_chem_comp_bond"

        if component.id == "D3O" and rem_hs:  # D3O has single heavy atom
            to_check.pop(2)

        assert cif_block
        assert component.id == cif_block.name
        for c in to_check:
            assert c in cif_block.get_mmcif_category_names()

    @staticmethod
    @pytest.mark.parametrize("rem_hs", [True, False])
    def test_plain_cif_write(component: Component, tmpdir, rem_hs):
        path = tmpdir.join(f"{component.id}.cif")
        to_check = must_have_categories.copy()

        component.ccd_cif_dict = None
        ccd_writer.write_molecule(str(path), component, remove_hs=rem_hs)
        cif_block = cif.read(str(path)).sole_block()

        if component.id == "NA":  # Na is an atom!
            to_check.pop(2)  # remove "_chem_comp_bond"

        if component.id == "D3O" and rem_hs:  # D3O has single heavy atom
            to_check.pop(2)

        assert cif_block
        assert component.id == cif_block.name
        for c in to_check:
            assert c in cif_block.get_mmcif_category_names()



# software from PDBe: Protein Data Bank in Europe; http://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Script for PDBeChem backend infrastructure.
Processes the wwPDB Chemical Components Dictionary file `components.cif`
producing files for:

http://ftp.ebi.ac.uk/pub/databases/msd/pdbechem_v2/

To do this components.cif is split into individual PDB chemical component
definitions cif files, sdf files, pdb files and image files.
In addition creates chem_comp.xml and chem_comp.list for all components.
More detailed description can be found here:

https://gitlab.ebi.ac.uk/pdbe/release/pdbechem
"""
import argparse
import logging
import os
import traceback
from pathlib import Path

import pdbeccdutils
import rdkit
from pdbeccdutils.core import ccd_reader, ccd_writer
from pdbeccdutils.core.component import Component
from pdbeccdutils.core.depictions import DepictionManager
from pdbeccdutils.core.exceptions import CCDUtilsError
from pdbeccdutils.core.fragment_library import FragmentLibrary
from pdbeccdutils.core.models import ConformerType, DepictionSource
from pdbeccdutils.utils import config
from pdbeccdutils.utils.pubchem_downloader import PubChemDownloader


def is_valid_path(path):
    p = Path(path)

    if not p.exists():
        raise argparse.ArgumentTypeError(f"{path} does not exists.")

    return p


class PDBeChemManager:
    """Manager orchestrating computation and generation of all parts of the
    PDBeChem update process.
    """

    def __init__(
        self,
        pubchem_templates="",
        general_templates=config.general_templates,
        library_path=config.fragment_library,
    ):
        """Initialize manager

        Args:
            pubchem_templates (str): Path to the pubchem templates
            general_templates (str, optional): Path to the general templates.
                Defaults to config.general_templates.
            library_path (str, optional): Path to the fragments library:
                Defaults to config.fragment_library.
        """
        # helper class to download templates if needed
        self.pubchem = (
            PubChemDownloader(pubchem_templates) if pubchem_templates else None
        )

        # helper class to get nice depictions
        self.depictions = DepictionManager(pubchem_templates, general_templates)

        # Fragments library to get substructure matches
        self.fragment_library = FragmentLibrary(library_path)

    def run(self, components_path, out_dir):
        """Process components

        Args:
            components_path (Path): Path to the components.cif file.
            out_dir (Path): Path to the out_dir
        """
        logging.info("Reading in components...")
        data = ccd_reader.read_pdb_components_file(components_path)

        for key, ccd_reader_result in data.items():
            ccd_out = out_dir / key[0] / key
            os.makedirs(ccd_out, exist_ok=True)
            self.process_single_component(ccd_reader_result, ccd_out)

            data[key] = None

        logging.debug("All is done!")

    def process_single_component(self, ccd_reader_result, out_dir):
        """Process single PDB-CCD component.

        Args:
            ccd_reader_result (CCDReaderResult): pdbeccdutils parser output.
            out_dir (Path): Out directory

        Return:
            bool: Whether or not all the files were succesfully written.
        """
        try:
            component = ccd_reader_result.component
            logging.info(f"{component.id} | processing...")

            # check parsing and conformer degeneration
            self._check_component_parsing(ccd_reader_result)
            self._generate_ideal_structure(component)

            # download templates if the user wants them.
            if self.pubchem is not None:
                self._download_template(component)

            # search fragment library
            self._search_fragment_library(component)

            # get scaffolds
            self._compute_component_scaffolds(component)

            # write out files
            self._generate_depictions(component, out_dir)
            self._export_structure_formats(component, out_dir)
            return True
        except Exception:
            logging.error(
                f"{ccd_reader_result.component.id} | FAILURE {traceback.format_exc()}."
            )
            return False

    def _check_component_parsing(self, ccd_reader_result):
        """Checks components parsing and highlights issues encountered with
        the molecule: errors/warnings during the parsing process,
        unrecoverable sanitization issues, inchikey mismatch between what
        was in the source file and is reproduced by rdkit.

        Args:
            ccd_reader_result (CCDReaderResult): Output of the parsing process.
        """

        if ccd_reader_result.warnings:
            logging.debug(f'warnings: {";".join(ccd_reader_result.warnings)}')

        if ccd_reader_result.errors:
            logging.debug(f'errors: {";".join(ccd_reader_result.errors)}')

        if not ccd_reader_result.sanitized:
            logging.debug("sanitization issue.")

        if not ccd_reader_result.component.inchikey_from_rdkit_matches_ccd():
            logging.debug("inchikey mismatch.")

    def _download_template(self, component: Component):
        """Attempts to download a pubchem template for the given component

        Args:
            component (Component): Component to be used.
        """
        component_downloaded = self.pubchem.process_template(component)
        if component_downloaded:
            logging.debug("downloaded new pubchem template.")

    def _generate_ideal_structure(self, component: Component):
        """Checks whether or not the component has degenerated ideal
        coordinates. If so, new conformer is attempted to be generated.

        Args:
            component (Component): Component to be
                processed.
        Return:
            bool: Whether the ideal coordinates have been successfully
            recalculated, false otherwise.
        """
        result = component.compute_3d()

        if component.has_degenerated_conformer(ConformerType.Ideal):
            logging.debug("has degenerated ideal coordinates.")

        if not result:
            logging.debug("error in generating 3D conformation.")

        return result

    def _search_fragment_library(self, component: Component):
        """Search fragment library to find hits

        Args:
            component (Component): Component to be processed
        """

        matches = component.library_search(self.fragment_library)

        if matches:
            logging.debug(
                f"{len(matches)} matches found in the library `{self.fragment_library.name}`."
            )

    def _compute_component_scaffolds(self, component: Component):
        """Compute scaffolds for a given component.

        Args:
            component (Component): Component to be processed
        """

        try:
            component.get_scaffolds()
        except CCDUtilsError as e:
            logging.error(str(e))

            return

        logging.debug(f"{len(component.scaffolds)} scaffold(s) were found.")

    def _generate_depictions(self, component: Component, out_dir: str):
        """Generate nice 2D depictions for the component and
        depiction annotations in JSON format. Presently depictions
        are generated in the following resolutions (100,200,300,400,500)
        with and without atom names.

        Args:
            component (Component): Component to be depicted.
            out_dir (str): Where the depictions should be stored.
        """
        depiction_result = component.compute_2d(self.depictions)

        if depiction_result.source == DepictionSource.Failed:
            logging.debug("failed to generate 2D image.")
        else:
            if depiction_result.score > 0.99:
                logging.debug("collision free image could not be generated.")
            logging.debug(
                f"2D generated using {depiction_result.source.name} with score {depiction_result.score}."
            )

        wedge_bonds = depiction_result.template_name != "cube"

        for i in range(100, 600, 100):
            component.export_2d_svg(
                os.path.join(out_dir, f"{component.id}_{i}.svg"),
                width=i,
                wedge_bonds=wedge_bonds,
            )
            component.export_2d_svg(
                os.path.join(out_dir, f"{component.id}_{i}_names.svg"),
                width=i,
                names=True,
                wedge_bonds=wedge_bonds,
            )

        component.export_2d_annotation(
            os.path.join(out_dir, f"{component.id}_annotation.json"),
            wedge_bonds=wedge_bonds,
        )

    def _export_structure_formats(self, component: Component, out_dir: Path):
        """Writes out component in a different formats as required for the
        PDBeChem FTP area.

        Args:
            component (Component): Component being processed.
            out_dir (Path): Where the results should be written
        """

        self.__write_molecule(
            out_dir / f"{component.id}_model.sdf",
            component,
            False,
            ConformerType.Model,
        )
        self.__write_molecule(
            out_dir / f"{component.id}_ideal.sdf",
            component,
            False,
            ConformerType.Ideal,
        )
        self.__write_molecule(
            out_dir / f"{component.id}_ideal_alt.pdb",
            component,
            True,
            ConformerType.Ideal,
        )
        self.__write_molecule(
            out_dir / f"{component.id}_model_alt.pdb",
            component,
            True,
            ConformerType.Model,
        )
        self.__write_molecule(
            out_dir / f"{component.id}_ideal.pdb",
            component,
            False,
            ConformerType.Ideal,
        )
        self.__write_molecule(
            out_dir / f"{component.id}_model.pdb",
            component,
            False,
            ConformerType.Model,
        )
        self.__write_molecule(
            out_dir / f"{component.id}.cml",
            component,
            False,
            ConformerType.Model,
        )
        self.__write_molecule(
            out_dir / f"{component.id}.cif",
            component,
            False,
            ConformerType.AllConformers,
        )

    def __write_molecule(self, path, component, alt_names, conformer_type):
        """Write out deemed structure.

        Args:
            path (Path): Path where the molecule will be stored.
            component (Component): Component to be written.
            alt_names (bool): Whether or not molecule will be written with
                alternate names.
            conformer_type (Component): Conformer to be written.
        """
        try:
            ccd_writer.write_molecule(
                path,
                component,
                remove_hs=False,
                alt_names=alt_names,
                conf_type=conformer_type,
            )
        except Exception:
            logging.error(f"error writing {path}.")

            with open(path, "w") as f:
                f.write("")


# region pre-light tasks
def create_parser():
    """
    Sets up parse the command line options.

    Returns:
         argparse.Namespace parser
    """
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "components_cif", help="Input PDB-CCD components.cif file (must be specified)"
    )
    parser.add_argument(
        "-g",
        "--general-templates",
        type=is_valid_path,
        default=config.general_templates,
        help="Use general templates in SDF format instead of those supplied with the code.",
    )
    parser.add_argument(
        "-p",
        "--pubchem-templates",
        type=is_valid_path,
        help="Path to the directory with pubchem templates in sdf format.",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        type=is_valid_path,
        required=True,
        help="Create an output directory with files suitable for PDBeChem ftp directory",
    )
    parser.add_argument(
        "-fl",
        "--fragment-library",
        type=is_valid_path,
        default=config.fragment_library,
        help="Use this fragment library in place of the one supplied with the code.",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Turn on debug message logging output"
    )

    return parser


# endregion


def main():
    """Runs the PDBeChem pipeline"""
    parser = create_parser()
    args = parser.parse_args()

    frm = "[%(asctime)-15s]  %(message)s"
    lvl = logging.DEBUG if __name__ == "__main__" and args.debug else logging.INFO

    logging.basicConfig(level=lvl, format=frm, datefmt="%a, %d %b %Y %H:%M:%S")
    logging.info("PDBeChem pipeline using:")
    logging.info(
        f"pdbeccdutils core v. {pdbeccdutils.__version__}, RDKit v. {rdkit.__version__}"
    )

    logging.info("Settings:")
    for k, v in vars(args).items():
        logging.info(f'{"":5s}{k:25s}{v}')

    pdbechem = PDBeChemManager(
        args.pubchem_templates, args.general_templates, args.fragment_library
    )
    pdbechem.run(args.components_cif, args.output_dir)


if __name__ == "__main__":
    main()
